/* 
 * $Id: bgp_util.c,v 1.31 1998/09/04 20:37:27 masaki Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <stdarg.h>
#include <netinet/in.h>
#include <unistd.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <setjmp.h>

#include <mrt.h>
#include <timer.h>
#include <trace.h>
#include <config_file.h>
#include <select.h>
#include <interface.h>
#include <rib.h>
#include <bgp.h>
#ifdef HAVE_IPV6
#include <api6.h>
#endif /* HAVE_IPV6 */


u_int aspath_hash_fn (aspath_t * aspath, u_int size);
void bgp_schedule_timer (mtimer_t * timer, bgp_peer_t * peer);
char *sbgp_states[] =
{
    "NULL",
    "Idle",
    "Connect",
    "Active",
    "Opensent",
    "Openconfirm",
    "Established"
};

char *sbgp_pdus[] =
{
    "NULL",
    "Open",
    "Update",
    "Notification",
    "Keepalive",
    "PACKET_MAX"
};

char *sbgp_events[] =
{
    "NULL",
    "Start",
    "Stop",
    "Open",
    "Closed",
    "OpenFail",
    "Error",
    "ConnectRetry",
    "HoldTime",
    "KeepAlive",
    "RecvOpen",
    "RecvKeepAlive",
    "RecvUpdate",
    "RecvNotify"
};


char *s_bgp_error[] =
{
    "UNUSED",
    "Message Header Error",
    "Open Message Error",
    "Update Message Error",
    "Hold mtimer_t Expired",
    "Finite State Machine Error",
    "Cease"
};

char *s_bgp_error_header[] =
{
    "UNUSED",
    "Connection Not Synchronized",
    "Bad Message Length",
    "Bad Message Type"
};

char *s_bgp_error_open[] =
{
    "UNUSED",
    "Unsupported Version Number",
    "Bad Peer AS",
    "Bad BGP Identifier",
    "Unsupported Authentication Code",
    "Authentication Failure",
    "Unacceptable Hold Time"
};

char *s_bgp_error_update[] =
{
    "Unused",
    "Malformed Attribute List",
    "Unrecognized Well-known Attribute",
    "Missing Well-known Attribute",
    "Attribute Flags Error",
    "Attribute Length Error",
    "Invalid ORIGIN Attribute",
    "AS Routing Loop",
    "Invalid NEXT_HOP Attribute",
    "Optional Attribute Error",
    "Invalid Network Field",
    "Malformed AS_PATH"
};


/* bgp_notify_to_string
 * return human readable string of BGP notify error
 */
char *
bgp_notify_to_string (int code, int subcode)
{
    char *tmpx;
    THREAD_SPECIFIC_STRAGE (tmpx);

    if (subcode > 0)
	switch (code) {
	case BGP_ERR_HEADER:
	    sprintf (tmpx, "%s (%s)", s_bgp_error[code],
		     s_bgp_error_header[subcode]);
	    break;
	case BGP_ERR_OPEN:
	    sprintf (tmpx, "%s (%s)", s_bgp_error[code],
		     s_bgp_error_open[subcode]);
	    break;
	case BGP_ERR_UPDATE:
	    sprintf (tmpx, "%s (%s)", s_bgp_error[code],
		     s_bgp_error_update[subcode]);
	    break;
	default:
	    sprintf (tmpx, "%s (No Subcode)", s_bgp_error[code]);
	}
    else
	sprintf (tmpx, "%s", s_bgp_error[code]);

    return (tmpx);
}


prefix_t *
get_local_addr (int sockfd)
{
    union sockaddr_any {
	struct sockaddr sa;
	struct sockaddr_in sin;
#ifdef HAVE_IPV6
	struct sockaddr_in6 sin6;
#endif /* HAVE_IPV6 */
    } addr;
    prefix_t *prefix = NULL;
    int len, family;

    len = sizeof (addr);
    memset (&addr, 0, sizeof (addr));
    if (getsockname (sockfd, (struct sockaddr *) &addr, &len) >= 0) {

	if ((family = addr.sa.sa_family) == AF_INET) {
	    struct sockaddr_in *sin = (struct sockaddr_in *) &addr;
	    prefix = New_Prefix (AF_INET, &sin->sin_addr, 32);
	}
#ifdef HAVE_IPV6
	else if (family == AF_INET6) {
	    struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *) &addr;
	    if (IN6_IS_ADDR_V4MAPPED (&sin6->sin6_addr))
		prefix = New_Prefix (AF_INET, ((u_char *) &sin6->sin6_addr) + 12, 32);
	    else
		prefix = New_Prefix (AF_INET6, &sin6->sin6_addr, 128);
	}
#endif /* HAVE_IPV6 */
	return (prefix);
    }
    return (NULL);
}


static void
peer_check_gateway (bgp_peer_t *peer)
{
    if (peer->local_addr)
	Deref_Prefix (peer->local_addr);
    peer->local_addr = get_local_addr (peer->sockfd);

    if (INTERFACE_MASTER && peer->gateway->interface == NULL) {
	prefix_t *prefix;

	peer->gateway->interface = find_interface_local (peer->local_addr);
	BIT_SET (peer->options, BGP_EBGP_MULTIHOP);
	prefix = rib_find_best_route (peer->gateway->prefix);
	if (prefix) {
	    if (peer->nexthop)
		Deref_Prefix (peer->nexthop);
	    peer->nexthop = Ref_Prefix (prefix);
	    trace (TR_INFO, peer->trace, "Multihop! nexthop will be %s\n",
	           prefix_toa (prefix));
	}
	else {
	    trace (TR_ERROR, peer->trace, "Multihop! nexthop unknown\n");
	}
    }
}


/* called from the select routine */
static void
bgp_connect_ready (bgp_peer_t * peer, int flag)
{
    struct sockaddr name;
    int namelen = sizeof (name);
#ifdef FIONBIO
    int optval = 0;
#endif /* FIONBIO */

    if (peer->sockfd < 0) {
	trace (TR_WARN, peer->trace,
	       "connect to %s succeeded but sockfd has been closed\n",
	       gateway_toa2 (peer->gateway));
	return;
    }

    if (peer->state != BGPSTATE_CONNECT) {
	trace (TR_WARN, peer->trace,
	       "connect to %s succeeded but too late\n",
	       gateway_toa2 (peer->gateway));
	return;
    }

    if (flag) select_delete_fd2 (peer->sockfd);

    /* see if we are really connected */
    if (getpeername (peer->sockfd, &name, &namelen) < 0) {
	trace (TR_INFO, peer->trace,
	       "connect to %s failed (%s)\n",
	       gateway_toa2 (peer->gateway), strerror (errno));
	close (peer->sockfd);
	peer->sockfd = -1;
	schedule_event2 ("bgp_sm_process_event",
			 peer->schedule, bgp_sm_process_event,
			 2, peer, BGPEVENT_OPENFAIL);
	return;
    }

    trace (TR_INFO, peer->trace, "connection to %s SUCCEEDED\n",
	   gateway_toa2 (peer->gateway));

#ifdef FIONBIO
    if (ioctl (peer->sockfd, FIONBIO, &optval) < 0) {
        trace (TR_ERROR, peer->trace, "ioctl FIONBIO failed (%s)\n",
               strerror (errno));
    }   
#else          
    if (fcntl (peer->sockfd, F_SETFL, 0) < 0) {
        trace (TR_ERROR, peer->trace, "fcntl F_SETFL 0 failed (%s)\n",
               strerror (errno));
    }   
#endif /* FIONBIO */     
    peer_check_gateway (peer);
    schedule_event2 ("bgp_sm_process_event",
		     peer->schedule, bgp_sm_process_event,
		     2, peer, BGPEVENT_OPEN);
}


static void
bgp_schedule_connect_ready (bgp_peer_t * peer)
{
    /* should be scheduled to avoid racing with incoming */
    schedule_event2 ("bgp_connect_ready",
		     peer->schedule, bgp_connect_ready, 2, peer, 1);
}


/* bgp_start_transport_connection
 * Starts tcp connection to peer. Returns 1 on sucess, -1 oltherwise
 */
int 
bgp_start_transport_connection (bgp_peer_t * peer)
{
    int optval = 1;
    int ret;
    int family, len;
    struct sockaddr_in sin;
    struct sockaddr *s;
#ifdef HAVE_IPV6
    struct sockaddr_in6 sin6;
#endif /* HAVE_IPV6 */

    /* okay, start a TCP connection */

    if ((family = peer->peer_addr->family) == AF_INET) {
	memset (&sin, 0, sizeof (sin));
	sin.sin_family = family;
	sin.sin_port = htons (peer->peer_port);
	memcpy (&sin.sin_addr, prefix_tochar (peer->peer_addr), 4);
	s = (struct sockaddr *) &sin;
	len = sizeof (sin);
    }

#ifdef HAVE_IPV6
    else if (family == AF_INET6) {
	memset (&sin6, 0, sizeof (sin6));
	sin6.sin6_family = family;
	sin6.sin6_port = htons (peer->peer_port);
	memcpy (&sin6.sin6_addr, prefix_tochar (peer->peer_addr), 16);
	s = (struct sockaddr *) &sin6;
	len = sizeof (sin6);
    }
#endif /* HAVE_IPV6 */
    else {
        return (-1);
    }

    if (peer->sockfd >= 0) {
	select_delete_fd (peer->sockfd);
    }

    if ((peer->sockfd = socket (family, SOCK_STREAM, 0)) < 0) {
	trace (TR_ERROR, peer->trace, "socket open failed (%s)\n",
	       strerror (errno));
	return (-1);
    }

    /* transport connection succeeded */
    setsockopt (peer->sockfd, SOL_SOCKET, SO_REUSEADDR,
		(char *) &optval, sizeof (optval));

    if (peer->bind_if) {
#if 0
	struct ifreq ifr;
	strncpy (&ifr.ifr_ifrn.ifrn_name, peer->bind_if->name, 
		 sizeof (ifr.ifr_ifrn.ifrn_name));
	if (setsockopt (peer->sockfd, SOL_SOCKET, SO_BINDTODEVICE,
				  &ifr, sizeof (ifr)) < 0)
	    trace (TR_ERROR, peer->trace, "SO_BINDTODEVICE %s (%s)\n",
				   peer->bind_if->name, strerror (errno));
#endif
    }

    if (peer->bind_addr) {
	    int llen;
	    struct sockaddr_in l_sin;
	    struct sockaddr *l_s;
#ifdef HAVE_IPV6
	    struct sockaddr_in6 l_sin6;
#endif /* HAVE_IPV6 */

	    if (family  == AF_INET) {
		    memset (&l_sin, 0, sizeof (sin));
		    l_sin.sin_family = family;
		    memcpy (&l_sin.sin_addr, prefix_tochar (peer->bind_addr), 
			    4);
		    l_s = (struct sockaddr *) &l_sin;
		    llen = sizeof (sin);
	    }
#ifdef HAVE_IPV6
	    else if (family == AF_INET6) {
		    memset (&l_sin6, 0, sizeof (l_sin6));
		    l_sin6.sin6_family = family;
		    memcpy (&l_sin6.sin6_addr, prefix_tochar (peer->bind_addr),
			    16);
		    l_s = (struct sockaddr *) &l_sin6;
		    llen = sizeof (sin6);
	    }
#endif /* HAVE_IPV6 */
	    else {
		assert (0);
		return (-1);
	    }
	    if (bind (peer->sockfd, l_s, len) < 0) {
		    trace (TR_ERROR, peer->trace, "socket bind (%s)\n",
			   strerror (errno));
		    return (-1);
	    }
    }


#ifndef HAVE_PTHREAD
#ifdef FIONBIO
    if (ioctl (peer->sockfd, FIONBIO, &optval) < 0) {
	trace (TR_ERROR, peer->trace, "ioctl FIONBIO failed (%s)\n",
	       strerror (errno));
    }
#else
    if (fcntl (peer->sockfd, F_SETFL, O_NONBLOCK) < 0) {
	trace (TR_ERROR, peer->trace, "fcntl F_SETFL O_NONBLOCK failed (%s)\n",
	       strerror (errno));
    }
#endif /* FIONBIO */
#endif /* HAVE_PTHREAD */

    ret = connect (peer->sockfd, s, len);
    if (ret < 0) {
	if (errno != EINPROGRESS) {
	    close (peer->sockfd);
	    peer->sockfd = -1;
	    return (-1);
	}
	trace (TR_PACKET, peer->trace, "waiting on %d for write\n",
	       peer->sockfd);
	select_add_fd (peer->sockfd, 2 /* write */, 
		       bgp_schedule_connect_ready, peer);
	return (0);
    }
    bgp_connect_ready (peer, 0);
    return (1);
}


static void
bgp_recv_open (bgp_peer_t * peer)
{
    int len;
    u_char *cp;
    int type, length;

    int version;
    int as;
    u_long holdtime;
    int authcode;
    u_long id;

    /* closed by outgoing */
    if (peer->sockfd_in < 0)
	return;

    len = bgp_read (peer, peer->sockfd_in, peer->read_ptr_in, BGPMAXPACKETSIZE);
    if (len < 0) {
        peer->read_ptr_in = peer->buffer_in; /* eat up the input */
    	select_delete_fd2 (peer->sockfd_in);
	close (peer->sockfd_in);
	peer->sockfd_in = -1;
        return;
    }

    peer->read_ptr_in += len;

    if ((len = peer->read_ptr_in - peer->buffer_in) < BGP_HEADER_LEN) {
	/* keep reading */
    	select_enable_fd (peer->sockfd_in);
	return;
    }

    cp = peer->buffer_in;

    BGP_GET_HDRTYPE (type, cp);
    BGP_GET_HDRLEN (length, cp);

    if (length < BGP_HEADER_LEN || length > BGPMAXPACKETSIZE) {
        select_delete_fd2 (peer->sockfd_in);
	trace (TR_ERROR, peer->trace,
	    "rejecting the incoming connection (bad length %d)\n", length);
        goto reject;
    }

    if (len < length) {
	/* keep reading */
    	select_enable_fd (peer->sockfd_in);
	return;
    }

    /* this happens only once */
    select_delete_fd2 (peer->sockfd_in);

    if (type >= BGP_OPEN && type < BGP_PACKET_MAX) {
        trace (TR_PACKET, peer->trace, "recv %s (%d bytes) [incoming]\n",
               sbgp_pdus[type], length);
    }
    else {
	trace (TR_ERROR, peer->trace,
	    "rejecting the incoming connection (unknown type code %d)\n", 
	    type);
	goto reject;
    }

    if (type != BGP_OPEN) {
	trace (TR_ERROR, peer->trace,
	    "rejecting the incoming connection (%s: type is not OPEN)\n",
	       sbgp_pdus[type]);
	goto reject;
    }

    BGP_SKIP_HEADER (cp);
    BGP_GET_OPEN_V4 (version, as, holdtime, id, authcode, cp);

    if (peer->state == BGPSTATE_ESTABLISHED ||
	peer->state == BGPSTATE_IDLE) {

	trace (TR_INFO, peer->trace,
	    "rejecting the incoming connection (state is %s)\n",
	       sbgp_states[peer->state]);
      reject:
	bgp_send_notification2 (peer->sockfd_in, peer->peer_addr, 
				peer->peer_port, BGP_CEASE, 0);
	close (peer->sockfd_in);
	peer->sockfd_in = -1;
	return;
    }

    if (peer->state == BGPSTATE_OPENSENT ||
	peer->state == BGPSTATE_OPENCONFIRM) {

	char his[MAXLINE], mine[MAXLINE];

	trace (TR_INFO, peer->trace,
	       "routerid comparison (in %s <=> out %s)\n", 
		inet_ntop (AF_INET, &id, his, sizeof (his)),
		inet_ntop (AF_INET, &BGP->my_id, mine, sizeof (mine)));
	if (ntohl (id) > ntohl (BGP->my_id)) {
	    goto accept;
	}
	else {
	    trace (TR_INFO, peer->trace,
	        "rejecting the incoming connection (router id)\n");
	    goto reject;
	}
    }

    if (peer->state == BGPSTATE_CONNECT ||
	peer->state == BGPSTATE_ACTIVE) {

      accept:
	trace (TR_INFO, peer->trace,
	       "BGP accepting the incoming connection (state was %s)\n",
	       sbgp_states[peer->state]);
	bgp_release_resources (peer);

	/* switch the inputs */
	peer->sockfd = peer->sockfd_in;
	peer->sockfd_in = -1;
	peer->packet = peer->buffer_in;

	/* if something remains */
	if (len > length) {
	    assert (peer->read_ptr == peer->start_ptr);
	    assert (peer->read_ptr == peer->buffer);
	    memcpy (peer->read_ptr, peer->buffer_in + length, len - length);
	    peer->read_ptr += (len - length);
	}

        peer_check_gateway (peer);

	if (bgp_process_open (peer) > 0) {
	    select_add_fd (peer->sockfd, 1, bgp_schedule_socket, peer);
	    bgp_change_state (peer, BGPSTATE_OPENSENT, BGPEVENT_OPEN);
	    if (bgp_send_open (peer) >= 0) {
	        bgp_change_state (peer, BGPSTATE_OPENCONFIRM, 
					BGPEVENT_RECVOPEN);
	        bgp_send_keepalive (peer);
	    }
	}
	/* OPEN was bad */
	else {
	    /* notification already sent */
	    /* outgoing connection has been closed. no way. */
	    bgp_change_state (peer, BGPSTATE_IDLE, BGPEVENT_RECVOPEN);
	    bgp_peer_dead (peer);
	}
	return;
    }
    /* all cases examined */
    assert (0);
}


static void
bgp_schedule_recv_open (bgp_peer_t * peer)
{
    peer->read_ptr_in = peer->buffer_in; /* reset the input */
    schedule_event2 ("bgp_recv_open",
		     peer->schedule, bgp_recv_open, 1, peer);
}


/* bgp_accept_connection

 */
static void
bgp_accept_connection (int fd)
{
    int sockfd, family, len;
    u_short port;
    bgp_peer_t *peer;
#ifndef HAVE_IPV6
    struct sockaddr_in addr, local;
#else
    union sockaddr_any {
	struct sockaddr sa;
	struct sockaddr_in sin;
	struct sockaddr_in6 sin6;
    } addr, local;
#endif /* HAVE_IPV6 */
    prefix_t *prefix, *local_prefix;

    len = sizeof (addr);

    if ((sockfd = accept (fd, (struct sockaddr *) &addr, &len)) < 0) {
	trace (TR_ERROR, BGP->trace, "accept (%s)\n", strerror (errno));
	select_enable_fd (fd);
	return;
    }

    len = sizeof (local);
    if (getsockname (sockfd, (struct sockaddr *) &local, &len) < 0) {
        trace (TR_ERROR, BGP->trace, "getsockname (%s)\n", strerror (errno));
        select_enable_fd (fd);
        return;
    }

    select_enable_fd (fd);

#ifndef HAVE_IPV6
    if ((family = addr.sin_family) == AF_INET) {
#else
    if ((family = addr.sa.sa_family) == AF_INET) {
#endif /* HAVE_IPV6 */
	struct sockaddr_in *sin = (struct sockaddr_in *) &addr;
	port = ntohs (sin->sin_port);
	prefix = New_Prefix (AF_INET, &sin->sin_addr, 32);
	sin = (struct sockaddr_in *) &local;
        local_prefix = New_Prefix (AF_INET, &sin->sin_addr, 32);
    }
#ifdef HAVE_IPV6
    else if (family == AF_INET6) {
	struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *) &addr;
	port = ntohs (sin6->sin6_port);
	if (IN6_IS_ADDR_V4MAPPED (&sin6->sin6_addr))
	    prefix = New_Prefix (AF_INET, ((char *) &sin6->sin6_addr) + 12, 
					  32);
	else
	    prefix = New_Prefix (AF_INET6, &sin6->sin6_addr, 128);
	sin6 = (struct sockaddr_in6 *) &local;
        if (IN6_IS_ADDR_V4MAPPED (&sin6->sin6_addr))
            local_prefix = New_Prefix (AF_INET, 
				      ((char *)&sin6->sin6_addr) + 12, 32);
        else
            local_prefix = New_Prefix (AF_INET6, &sin6->sin6_addr, 128);
    }
#endif /* HAVE_IPV6 */
    else {
	trace (TR_ERROR, BGP->trace, "unknown connection family = %d\n",
	       family);
	close (sockfd);
	return;
    }

    peer = Find_BGP_Peer (prefix, local_prefix, NULL);

    if (peer) {
	trace (TR_INFO, peer->trace, "Valid Connection Accepted on %s\n",
		prefix_toa (local_prefix));
    }
    else {
	if (BGP->accept_all_peers) {
	    /* at this point, as is unknown */
	    /* modifying BGP structure should be safe */
	    peer = Add_BGP_Peer (NULL, prefix, local_prefix, NULL, 0, 
				 BGP->trace);
	    /* safer way */
            BIT_SET (peer->options, BGP_CONNECT_PASSIVE);
	    trace (TR_INFO, peer->trace, "Creating Connection on %s\n",
		   prefix_toa (local_prefix));
	    /* force to change BGP state */
	    bgp_change_state (peer, BGPSTATE_ACTIVE, BGPEVENT_START);
	}
	else {
	    trace (TR_INFO, BGP->trace, "Unconfigured peer %s on %s\n",
		   prefix_toa (prefix), prefix_toa (local_prefix));
	    bgp_send_notification2 (sockfd, prefix, ntohs (port),
				   BGP_ERR_OPEN, BGP_ERROPN_AUTH);
	    Deref_Prefix (prefix);
	    Deref_Prefix (local_prefix);
	    close (sockfd);
	    return;
	}
    }

    /* XXX I think we need to accept all incoming connections here,
       and later determine if it's really accepted or not 
       after receiving an OPEN message through the connection. */

/* accept even in idle state for the time being -- masaki */
#if 0
    if (peer->state == BGPSTATE_IDLE)
	peer->state = BGPSTATE_ACTIVE;
#endif
    if (peer->state == BGPSTATE_IDLE ||
	peer->state == BGPSTATE_ESTABLISHED) {
	trace (TR_INFO, peer->trace, "rejecting the connection (%s)\n",
	       sbgp_states[peer->state]);
	bgp_send_notification2 (sockfd, prefix, ntohs (port), BGP_CEASE, 0);
	Deref_Prefix (prefix);
        Deref_Prefix (local_prefix);
	close (sockfd);
	return;
    }

    Deref_Prefix (prefix);
    Deref_Prefix (local_prefix);

    if (peer->sockfd_in >= 0)
	select_delete_fd (peer->sockfd_in);
    peer->sockfd_in = sockfd;
    peer->read_ptr_in = peer->buffer_in;
    select_add_fd (sockfd, 1, bgp_schedule_recv_open, peer);
}



/*-----------------------------------------------------------
 *  Name: 	bgp_release_resources
 *  Created:	Fri Dec 23 14:16:27 1994
 *  Author: 	Craig Labovitz   <labovit@snoopy.merit.net>
 *  DESCR:  	
 */

void 
bgp_release_resources (bgp_peer_t * peer)
{

    if (peer->sockfd >= 0) {
	if (peer->sockfd >= 0) {
	    trace (TR_INFO, peer->trace, "outgoing connection closed\n");
	}
	select_delete_fd (peer->sockfd);
	peer->sockfd = -1;
    }

    Timer_Turn_OFF (peer->timer_ConnectRetry);
    Timer_Turn_OFF (peer->timer_KeepAlive);
    Timer_Turn_OFF (peer->timer_HoldTime);

    peer->read_ptr = peer->buffer;
    peer->start_ptr = peer->buffer;
    peer->packet = NULL;
    peer->read_ptr_in = peer->buffer_in;

    /* should flush input buffer here */
    /* delete sockfd stuff */
    clear_schedule (peer->schedule);

}


u_int 
aspath_lookup_fn (aspath_t * tmp1, aspath_t * tmp2)
{
    u_int tmpx;

    if ((tmp1 == NULL) || (tmp2 == NULL)) {
	return (-1);
    }

    tmpx = compare_aspaths (tmp1, tmp2);
    return (tmpx);
}


/* init_BP
 */
int 
init_BGP (trace_t * ltrace)
{
    bgp_attr_t attr;
    char str[MAXLINE];

    BGP = New (bgp_t);
    BGP->ll_bgp_peers = LL_Create (0);
    BGP->ll_networks = LL_Create (0);

    BGP->lport = BGP_PORT;
    BGP->cport = BGP_PORT;
    BGP->sockfd = -1;

    BGP->accept_all_peers = 0;
    BGP->my_as = /* DEFAULT_MY_AS */ 0 /* must be set */;
    BGP->my_id = /* htonl (DEFAULT_MY_ID) */ 0 /* must be set */;
    BGP->trace = trace_copy (ltrace);
    set_trace (BGP->trace, TRACE_PREPEND_STRING, "BGP", 0);
    BGP->views[0] = New_View (32);
    BGP->views[0]->viewno = 0;
    BGP->views[0]->trace = trace_copy (ltrace);
    BGP->views[0]->rib = 1;
    sprintf (str, "BGP view %d", BGP->views[0]->viewno);
    set_trace (BGP->views[0]->trace, TRACE_PREPEND_STRING, str, 0);
#ifdef HAVE_IPV6
    BGP->views[1] = New_View (128);
    BGP->views[1]->viewno = 1;
    BGP->views[1]->trace = trace_copy (ltrace);
    BGP->views[1]->rib = 1;
    sprintf (str, "BGP4+ view %d", BGP->views[1]->viewno);
    set_trace (BGP->views[1]->trace, TRACE_PREPEND_STRING, str, 0);
#endif /* HAVE_IPV6 */
    BGP->default_local_pref = DEFAULT_LOCAL_PREF;
    BGP->bgp_num_active_route_head = 0;
    BGP->bgp_num_active_route_node = 0;
    BGP->Default_Start_Interval = DEFAULT_START_INTERVAL;
    BGP->Default_ConnectRetry_Interval = DEFAULT_CONNET_RETRY_INTERVAL;
    BGP->Default_KeepAlive_Interval = DEFAULT_KEEPALIVE_INTERVAL;
    BGP->Default_HoldTime_Interval = DEFAULT_HOLDTIME_INTERVAL;

    pthread_mutex_init (&BGP->mutex_lock, NULL);

    BGP->attr_hash =
	HASH_Create (100, HASH_KeyOffset, HASH_Offset (&attr, &attr.aspath),
		     HASH_LookupFunction, aspath_lookup_fn,
		     HASH_HashFunction, aspath_hash_fn, NULL);

    BGP->schedule = (schedule_t *) New_Schedule ("BGP", BGP->trace);
    mrt_thread_create ("BGP thread", BGP->schedule,
		       (thread_fn_t) bgp_start_main_thread, 0);

    return (1);
}


/* 
 * Start BGP listening on port (usually 179)
 */
static int 
init_BGP_listen (bgp_peer_t *peer)
{

    struct sockaddr *sa;
    int sockfd, len, optval = 1;

    struct sockaddr_in serv_addr;
#ifdef HAVE_IPV6
    struct sockaddr_in6 serv_addr6;
#endif /* HAVE_IPV6 */

    assert (peer == NULL || peer->bind_addr || peer->bind_if);

#ifdef HAVE_IPV6
    memset (&serv_addr6, 0, sizeof (serv_addr6));
    serv_addr6.sin6_port = htons (BGP->lport);
    serv_addr6.sin6_family = AF_INET6;
    if (peer && peer->bind_addr && peer->bind_addr->family == AF_INET6)
        memcpy (&serv_addr6.sin6_addr, prefix_tochar (peer->bind_addr), 16);
#endif /* HAVE_IPV6 */

    /* check that NOT listening to portmaster! */
    memset (&serv_addr, 0, sizeof (serv_addr));
    serv_addr.sin_port = htons (BGP->lport);
    serv_addr.sin_family = AF_INET;
    if (peer && peer->bind_addr && peer->bind_addr->family == AF_INET)
        memcpy (&serv_addr.sin_addr, prefix_tochar (peer->bind_addr), 4);

#ifdef HAVE_IPV6
    sa = (struct sockaddr *) &serv_addr6;
    len = sizeof (serv_addr6);
#else
    sa = (struct sockaddr *) &serv_addr;
    len = sizeof (serv_addr);
#endif /* HAVE_IPV6 */

    if ((sockfd = socket (sa->sa_family, SOCK_STREAM, 0)) < 0) {
#ifdef HAVE_IPV6
	if (sa->sa_family == AF_INET)
	    goto error;
	sa = (struct sockaddr *) &serv_addr;
	len = sizeof (serv_addr);
        if ((sockfd = socket (sa->sa_family, SOCK_STREAM, 0)) < 0) {
error:
#endif /* HAVE_IPV6 */
	    trace (TR_ERROR, BGP->trace, 
		   "Could not get socket (%s)\n", strerror (errno));
	    return (-1);
#ifdef HAVE_IPV6
	}
#endif /* HAVE_IPV6 */
    }

    /* allow the reuse this port */
    if (setsockopt (sockfd, SOL_SOCKET, SO_REUSEADDR,
		    (char *) &optval, sizeof (optval)) < 0)
	trace (TR_ERROR, BGP->trace, "setsockopt failed (%s)\n",
	       strerror (errno));

#ifdef SO_BINDTODEVICE
    if (peer && peer->bind_if) {
            struct ifreq ifr;
            strncpy (ifr.ifr_ifrn.ifrn_name, peer->bind_if->name, 
			sizeof (ifr.ifr_ifrn.ifrn_name));
            if (setsockopt (sockfd, SOL_SOCKET, SO_BINDTODEVICE,
                                  &ifr, sizeof (ifr)) < 0)
                 trace (TR_ERROR, peer->trace, "BINDTODEVICE %s (%s)\n",
                                   peer->bind_if->name, strerror (errno));
	    else
                 trace (TR_INFO, BGP->trace, "binding to interface %s\n",
			peer->bind_if->name);
    }
#endif /* SO_BINDTODEVICE */

    if (bind (sockfd, sa, len) < 0) {
	trace (TR_ERROR, BGP->trace, 
	       "Could not bind to port %d (%s), aborting\n",
	       BGP->lport, strerror (errno));
	close (sockfd);
	return (-1);
    }

    listen (sockfd, 5);

    trace (TR_INFO, BGP->trace, "listening for connections on port %d\n",
	   BGP->lport);

    if (peer)
	peer->sockfd_listen = sockfd;
    else
        BGP->sockfd = sockfd;
    return (1);
}


bgp_peer_t *
Find_BGP_Peer (prefix_t * prefix, prefix_t * usrc, interface_t *interface)
{
    int score = -1;
    bgp_peer_t *best = NULL;
    bgp_peer_t *peer;

    /* scan through peers */
    LL_Iterate (BGP->ll_bgp_peers, peer) {
	    int quality = 0;

	if (prefix_compare (peer->peer_addr, prefix))
		goto chk_2;

	if (peer->aliases) {
	    prefix_t *alias;

	    LL_Iterate (peer->aliases, alias) {
		if (prefix_compare (alias, prefix))
		    goto chk_2;
	    }
	}
	continue;

chk_2:
	if (interface && usrc) {
		if (peer->bind_if && peer->bind_if != interface)
			continue;
		if (peer->bind_addr &&
		    !prefix_compare (peer->bind_addr, usrc))
			continue;
		if (peer->bind_addr)
			quality |= 2;
		if (peer->bind_if)
			quality |= 1;
	} else if (usrc) {
		if (peer->bind_addr &&
		    !prefix_compare (peer->bind_addr, usrc)) {
			trace (TR_ERROR, peer->trace, "not matched usrc=%s\n",
			       prefix_toa (usrc));
			continue;
		}
		if (peer->bind_addr)
			quality |= 2;
		if (!peer->bind_if)
			quality |= 1;
	} else if (interface) {
		if (peer->bind_if && peer->bind_if != interface)
			continue;
		if (!peer->bind_addr)
			quality |= 2;
		if (peer->bind_if)
			quality |= 1;
	} else {
		if (!peer->bind_addr)
			quality |= 2;
		if (!peer->bind_if)
			quality |= 1;
	}
	    if (quality > score) {
		    best = peer;
		    score = quality;
	    }
    }
    return (best);
}


bgp_peer_t *
Find_BGP_Peer_ByID (char *name)
{
    bgp_peer_t *peer;

    /* scan through peers */
    LL_Iterate (BGP->ll_bgp_peers, peer) {
	    if (peer->name && strcasecmp (peer->name, name) == 0)
		    return (peer);
    }
    return (NULL);
}


static u_long
bgp_obtain_bit ()
{
    int i;

    for (i = 0 ; i < MAX_BGP_PEERS; i++) {
        if ((BGP->bits & (1 << i)) == 0) {
            BGP->bits |= (1 << i);
	    return (1 << i);
        }
    }
    return (0);
}

static void
bgp_return_bit (u_long bit)
{
    assert ((BGP->bits & bit) != 0);
    BGP->bits &= ~bit;
}


/* 
 * Initialize BGP peer and add to list of peers
 */
bgp_peer_t *
Add_BGP_Peer (char *name, prefix_t * prefix, 
	      prefix_t *usrc, interface_t *interface, int as, trace_t * trace)
{
    char str[MAXLINE];
    bgp_peer_t *peer;

    pthread_mutex_lock (&BGP->mutex_lock);
    if ((peer = Find_BGP_Peer (prefix, usrc, interface)) != NULL) {
        pthread_mutex_unlock (&BGP->mutex_lock);
	return (NULL);
    }
#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
        sprintf (str, "BGP4+ %s", (name)? name: prefix_toa (prefix));
    else
#endif /* HAVE_IPV6 */
    sprintf (str, "BGP %s", (name)? name: prefix_toa (prefix));

    peer = New (bgp_peer_t);
    peer->name = (name)? strdup (name): NULL;
    peer->trace = trace_copy (trace);
    set_trace (peer->trace, TRACE_PREPEND_STRING, str, 0);
#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
	peer->view = 1;
    else
#endif /* HAVE_IPV6 */
    { peer->view = 0; BGP->num_ipv4_peers++; }
    peer->bit = bgp_obtain_bit ();
    BGP->num_peers++;

    peer->gateway = add_gateway (prefix, as, NULL);
    peer->peer_addr = Ref_Prefix (prefix);
    peer->bind_addr = Ref_Prefix (usrc);
    peer->bind_if = interface;
    peer->peer_port = BGP->cport;	/* usually 179 */
    peer->peer_id = 0L;
    peer->aliases = NULL;
    peer->sockfd = -1;
    peer->sockfd_in = -1;
    peer->sockfd_listen = -1;

    peer->dlist_in = peer->dlist_out = -1;
    peer->flist_in = peer->flist_out = -1;
    peer->route_map_in = peer->route_map_out = -1;

    peer->read_ptr = peer->buffer;
    peer->start_ptr = peer->buffer;
    peer->packet = NULL;
    peer->read_ptr_in = peer->buffer_in;

    peer->state = BGPSTATE_IDLE;

    peer->maximum_prefix = 0;

    peer->Start_Interval = -1;
    peer->ConnectRetry_Interval = -1;
    peer->KeepAlive_Interval = -1;
    peer->HoldTime_Interval = -1;

    /* create timers */
    peer->timer_ConnectRetry = New_Timer (bgp_schedule_timer,
					 BGP->Default_ConnectRetry_Interval,
					 "ConnectRetry", peer);
    peer->timer_KeepAlive = New_Timer (bgp_schedule_timer,
				      BGP->Default_KeepAlive_Interval,
				      "KeepALive", peer);
    peer->timer_HoldTime = New_Timer (bgp_schedule_timer,
				     BGP->Default_HoldTime_Interval,
				     "HoldTime", peer);
    peer->timer_Start = New_Timer (bgp_schedule_timer, 
				   BGP->Default_Start_Interval,
				  "StartTime", peer);
    peer->schedule = New_Schedule (gateway_toa2 (peer->gateway), peer->trace);

    LL_Add (BGP->ll_bgp_peers, peer);

    pthread_mutex_unlock (&BGP->mutex_lock);

    mrt_thread_create (str, peer->schedule,
		       (thread_fn_t) bgp_start_peer_thread, peer);
    return (peer);
}


static void
bgp_destroy_peer (bgp_peer_t *peer)
{
    bgp_peer_t *bp;

    pthread_mutex_lock (&BGP->mutex_lock);

    LL_Iterate (BGP->ll_bgp_peers, bp) {
	if (bp == peer)
	    break;
    }

    /* this is done asynchronously, so it may happen */
    if (bp == NULL) {
        pthread_mutex_unlock (&BGP->mutex_lock);
	return;
    }

    LL_Remove (BGP->ll_bgp_peers, peer);
#ifdef HAVE_IPV6
    if (peer->peer_addr->family != AF_INET6)
#endif /* HAVE_IPV6 */
    BGP->num_ipv4_peers++;
    BGP->num_peers--;
    bgp_return_bit (peer->bit);

    pthread_mutex_unlock (&BGP->mutex_lock);

    /* peer is not on the list now, but the routunes should not care */
    bgp_peer_dead (peer);

    Deref_Prefix (peer->peer_addr);
    if (peer->name) Delete (peer->name);
    Destroy_Timer (peer->timer_ConnectRetry);
    Destroy_Timer (peer->timer_KeepAlive);
    Destroy_Timer (peer->timer_HoldTime);
    Destroy_Timer (peer->timer_Start);
    Destroy_Trace (peer->trace);
    delete_schedule (peer->schedule);

    mrt_thread_exit ();
}


void
Destroy_BGP_Peer (bgp_peer_t *peer)
{
    schedule_event2 ("bgp_destroy_peer",
		      peer->schedule, bgp_destroy_peer, 1, peer);
}


/* bgp_peer_dead
 * peer has left established state, withdraw routes and
 * free resources
 */
void 
bgp_peer_dead (bgp_peer_t * peer)
{
    int i;

    trace (TR_INFO, peer->trace, "dead -- closing connection\n");

    /*bgp_change_state (peer, BGPSTATE_IDLE, BGPEVENT_ERROR);   */
    for (i = 0; i < MAX_BGP_VIEWS; i++) {
	if (BGP->views[i])
	    view_delete_peer (BGP->views[i], peer);
    }
    bgp_release_resources (peer);

    if (BGP->peer_down_call_fn)
	BGP->peer_down_call_fn (peer);
    else
	default_bgp_peer_down_fd (peer);
}


int
bgp_kill_peer (gateway_t *gateway)
{
    bgp_peer_t *peer;

    if ((peer = find_bgp_peer (gateway)) != NULL) {
	bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (1);
    }
    return (-1);
}


int
bgp_kill_peer2 (bgp_peer_t *peer)
{
    assert (peer);
    schedule_event2 ("bgp_sm_process_event",
		      peer->schedule, bgp_sm_process_event,
		      2, peer, BGPEVENT_ERROR);
    return (1);
}


void
bgp_kill_all ()
{
    bgp_peer_t *peer;

    pthread_mutex_lock (&BGP->mutex_lock);
    LL_Iterate (BGP->ll_bgp_peers, peer) {
        bgp_kill_peer2 (peer);
    }
    pthread_mutex_unlock (&BGP->mutex_lock);
}


typedef void (*void_func_t)();

/* set_BGP
 */
int 
set_BGP (int first, ...)
{
    va_list ap;
    enum BGP_ATTR attr;
    int id;

    assert (BGP);
    pthread_mutex_lock (&BGP->mutex_lock);
    /* Process the Arguments */
    va_start (ap, first);
    for (attr = (enum BGP_ATTR) first; attr; 
				       attr = va_arg (ap, enum BGP_ATTR)) {
	switch (attr) {
	case BGP_MY_AS:
	    BGP->my_as = va_arg (ap, int);
	    break;
	case BGP_MY_ID:
	    BGP->my_id = va_arg (ap, int);
	    break;
	case BGP_PEER_DOWN_FN:
	    BGP->peer_down_call_fn = va_arg (ap, void_func_t);
	    break;
	case BGP_PEER_ESTABLISHED_FN:
	    BGP->peer_established_call_fn = va_arg (ap, void_func_t);
	    break;
	case BGP_RECV_UPDATE_FN:
	    BGP->update_call_fn = va_arg (ap, void_func_t);
	    break;
	case BGP_RT_UPDATE_FN:
	    BGP->rt_update_call_fn = va_arg (ap, rib_update_route_t);
	    break;
	case BGP_STATE_CHANGE_FN:
	    BGP->state_change_fn = va_arg (ap, void_func_t);
	    break;
	case BGP_ACCEPT_ALL_PEERS:
	    BGP->accept_all_peers = va_arg (ap, int);
	    break;
	case BGP_LPORT:
	    BGP->lport = va_arg (ap, int);
	    break;
	case BGP_CPORT:
	    BGP->cport = va_arg (ap, int);
	    break;
	case BGP_TRACE_STRUCT:
	    BGP->trace = va_arg (ap, trace_t *);
	    break;
	case BGP_DUMP_ROUTE_FORM:
	    id = va_arg (ap, int);
	    assert (id >= 0 && id < MAX_BGP_VIEWS);
	    if (BGP->dump_route_form[id])
		free (BGP->dump_route_form[id]);
	    BGP->dump_route_form[id] = strdup (va_arg (ap, char *));
	    BGP->dump_route_interval[id] = va_arg (ap, int);
	    BGP->dump_route_type = va_arg (ap, int);
	    BGP->dump_route_time[id] = 0;
	    break;
	case BGP_DUMP_UPDATE_FORM:
	    if (BGP->dump_update_form)
		free (BGP->dump_update_form);
	    BGP->dump_update_form = strdup (va_arg (ap, char *));
	    BGP->dump_update_interval = va_arg (ap, int);
	    BGP->dump_update_time = 0;
	    break;
	default:
	    assert (0);
	    break;
	}
    }
    va_end (ap);
    pthread_mutex_unlock (&BGP->mutex_lock);

    return (1);
}


int 
default_bgp_peer_down_fd (bgp_peer_t * peer)
{

    if (!BIT_TEST (BGP_CONNECT_PASSIVE, peer->options))
	Timer_Turn_ON (peer->timer_Start);
    return (1);
}


/* find_bgp_peer
 * given a prefix, find the corresponding bgp peer
 * structure.
 */
bgp_peer_t *
find_bgp_peer (gateway_t * gateway)
{
    bgp_peer_t *peer;

    LL_Iterate (BGP->ll_bgp_peers, peer) {
	if (peer->gateway == gateway)
	    return (peer);
    }

    return (NULL);
}


void
start_bgp_peer (bgp_peer_t *peer)
{
    int doit = 0;

/* XXX not implemented listening on a specific address and interface */
    if (BGP->sockfd < 0) {
        if (init_BGP_listen (NULL) < 0)
	    return;
	doit++;
    }

    if (BIT_TEST (BGP_CONNECT_PASSIVE, peer->options)) {
	bgp_change_state (peer, BGPSTATE_ACTIVE, BGPEVENT_START);
    }
    else {
        schedule_event2 ("bgp_sm_process_event",
		          peer->schedule, bgp_sm_process_event,
		          2, peer, BGPEVENT_START);
    }
    if (doit)
        select_add_fd (BGP->sockfd, 1, bgp_accept_connection, 
		       (void *) BGP->sockfd);
}


/*
 * put all peers into start state, and begin listening on BGP
 * port (usually 179)
 */
int 
start_bgp ()
{
    bgp_peer_t *peer;

    LL_Iterate (BGP->ll_bgp_peers, peer) {
	start_bgp_peer (peer);
    }
    return (1);
}


/* 
 * dump various BGP stats to a socket
 * usually called by UII (user interactive interface)
 */
int 
show_bgp (uii_connection_t * uii)
{
    bgp_peer_t *peer;
    char tmpx[MAXLINE], date[MAXLINE];
    time_t now, t;

    pthread_mutex_lock (&BGP->mutex_lock);
    time (&now);
#ifdef HAVE_IPV6
    uii_send_data (uii, "Routing Protocol is \"BGP4+\"\n");
#else
    uii_send_data (uii, "Routing Protocol is \"BGP\"\n");
#endif /* HAVE_IPV6 */
    uii_send_data (uii, "Local Router ID is %s\n", 
		   inet_ntop (AF_INET, &BGP->my_id, tmpx, sizeof (tmpx)));
    uii_send_data (uii, "Local AS is %d\n", BGP->my_as);
    if (BGP->trace != NULL)
	uii_send_data (uii, "Trace flags 0x%x\n", BGP->trace->flags);
    /* uii_send_data (uii, "%d Attribute Blocks\n", 
	              HASH_GetCount (BGP->attr_hash)); */
    if (BGP->bgp_num_active_route_head > 0)
        uii_send_data (uii, "%5d bgp active route heads\n",
		       BGP->bgp_num_active_route_head);
    if (BGP->bgp_num_active_route_node > 0)
        uii_send_data (uii, "%5d bgp active route nodes\n",
		       BGP->bgp_num_active_route_node);

    LL_Iterate (BGP->ll_bgp_peers, peer) {
	char xxx[MAXLINE];
#ifdef HAVE_IPV6
	int bgp4plus_version = 0;
#endif /* HAVE_IPV6 */

	t = now - peer->time;
	if (t / 3600 > 99)
	    sprintf (date, "%02lddy%02ldhr", 
			   t / (3600 * 24), (t % (3600 * 24)) / 3600);
	else
	    sprintf (date, "%02ld:%02ld:%02ld", 
			   t / 3600, (t / 60) % 60, t % 60);

	xxx[0] = '\0';
#ifdef HAVE_IPV6
	if (BIT_TEST (peer->options, BGP_BGP4PLUS_01))
	    bgp4plus_version = 1;
	if (peer->gateway->prefix->family == AF_INET6)
	    sprintf (xxx, "BGP4+ draft %d", bgp4plus_version);
#endif /* HAVE_IPV6 */

	uii_send_data (uii, "\n");
        if (peer->name)
	    sprintf (tmpx, "%s (%s)", peer->name, prefix_toa (peer->peer_addr));
	else
	    sprintf (tmpx, "%s", prefix_toa (peer->peer_addr));
	uii_send_data (uii, "  peer %s AS%d on %s [%s] %s %s\n", tmpx,
		       peer->gateway->AS, peer->gateway->interface->name,
		       sbgp_states[peer->state], date, xxx);

	uii_send_data (uii, 
		 "    Router ID %s (BIT 0x%x) View #%d weight %d\n",
		 inet_ntop (AF_INET, &peer->gateway->routerid, tmpx, 
			    sizeof (tmpx)),
		 peer->bit, peer->view, peer->default_pref);

	if (peer->timer_KeepAlive->on != 1)
	    uii_send_data (uii, "    KeepAlive Off");
	else
	    uii_send_data (uii, "    KeepAlive %d",
			   peer->timer_KeepAlive->time_next_fire - now);
	if (peer->timer_Start->on != 1)
	    uii_send_data (uii, "  Starttimer Off");
	else
	    uii_send_data (uii, "  Starttimer %d",
			   peer->timer_Start->time_next_fire - now);
	if (peer->timer_HoldTime->on != 1)
	    uii_send_data (uii, "  Holdtime Off");
	else
	    uii_send_data (uii, "  Holdtime %d",
			   peer->timer_HoldTime->time_next_fire - now);
	if (peer->timer_ConnectRetry->on != 1)
	    uii_send_data (uii, "  ConnectRetry Off\n");
	else
	    uii_send_data (uii, "  ConnectRetry %d\n",
			   peer->timer_ConnectRetry->time_next_fire - now);
    }
    pthread_mutex_unlock (&BGP->mutex_lock);

    return (1);
}

/* bgp_debug
 * set the debug flags or file for BGP
 * called by config routines
 */
int 
bgp_debug (uii_connection_t * uii, int opt_num, char *s)
{
    char *token, *arg = s;

    pthread_mutex_lock (&BGP->mutex_lock);
    /* get flag */
    if (opt_num == 0 || (token = uii_parse_line (&s)) == NULL) {
        set_trace (BGP->trace, TRACE_FLAGS, TR_ALL, NULL);
    }
    else {
        u_long flag;

        if ((flag = trace_flag (token)) == 0) {
            config_notice (TR_ERROR, uii, "unknown debug option\n");
	    Delete (arg);
	    return (-1);
	}
        set_trace (BGP->trace, TRACE_FLAGS, flag, NULL);

        /* get file name (OPTIONAL) */
        if ((token = uii_parse_line (&s)) == NULL) {
	    Delete (arg);
	    return (1);
        }
        set_trace (BGP->trace, TRACE_LOGFILE, token, NULL);
    }
    pthread_mutex_unlock (&BGP->mutex_lock);
    return (1);
}


int 
trace_bgp (uii_connection_t * uii)
{
    pthread_mutex_lock (&BGP->mutex_lock);
    set_trace (BGP->trace, TRACE_ADD_FLAGS, TR_ALL, NULL);
    pthread_mutex_unlock (&BGP->mutex_lock);
    return (1);
}


int 
no_trace_bgp (uii_connection_t * uii)
{
    pthread_mutex_lock (&BGP->mutex_lock);
    set_trace (BGP->trace, TRACE_DEL_FLAGS, TR_ALL, NULL);
    pthread_mutex_unlock (&BGP->mutex_lock);
    return (1);
}


/* bgp_dump_view
 * dump BGP routes to socket. usually called by UII
 */
void
bgp_dump_view (uii_connection_t * uii, view_t *view,
		as_regexp_code_t *code, bgp_peer_t *peer)
{
    bgp_route_head_t *rt_head;
    bgp_route_t *route;
    char *s_origin[] = {"i", "e", "?", "a"};
    char tmpx[MAXLINE];
    int t, c1, c2;
    time_t now;

    if (peer) {
        view = BGP->views[peer->view];
        uii_send_data (uii, "Peer %s View %d\n", 
		       prefix_toa (peer->peer_addr), view->viewno);
    }
    else {
        uii_send_data (uii, "View #%d %s %d routes\n",
		   view->viewno, family2string (view->family), 
		   view->num_active_routes);
    }
    uii_send_data (uii,
	    "Status code: s suppressed, * valid, > best, i - internal\n");
    uii_send_data (uii,
	    "Origin codes: i - IGP, e - EGP, ? - incomplete, a - aggregate\n");
    uii_send_data (uii, "\n");
    rib_show_route_head (uii, "Path");

    time (&now);

    VIEW_RADIX_WALK (view, rt_head) {

if (peer) {
	    /* show only active one */
	    route = (rt_head->imported)?rt_head->imported:rt_head->active;

	    if (BIT_TEST (rt_head->peer_mask, peer->bit)) {

	        t = now - route->time;
	        c1 = '>';
	        if (BIT_TEST (rt_head->state_bits, VRTS_SUPPRESS))
		    c1 = 's';
	        if (BIT_TEST (rt_head->state_bits, VRTS_DELETE))
		    c1 = 'D';

	        strcpy (tmpx, "");
	        if (BGP4_BIT_TEST (route->attr->attribs, PA4_TYPE_ASPATH)) {
		    strcat (tmpx, aspath_toa (route->attr->aspath));
	            strcat (tmpx, " ");
	        }

	        if (BGP4_BIT_TEST (route->attr->attribs, PA4_TYPE_ORIGIN)) {
		    strcat (tmpx, s_origin[route->attr->origin]);
	            strcat (tmpx, " ");
	        }
    
		c2 = (route->attr->gateway->AS == BGP->my_as)? 'i': ' ';
		c2 = (route->aggregated)? 'a': c2;
	        rib_show_route_line (uii, c1, c2, route->attr->type, 
				     route->pref, t,
			             rt_head->prefix, 
				     route->attr->nexthop,
/* (route->attr->link_local)? route->attr->link_local: route->attr->nexthop, */
				     route->attr->gateway, tmpx);
	    }
}
else {

#ifdef notdef
	/* can not use continue in RADIX_WALK macro */
	if (BIT_TEST (rt_head->state_bits, VRTS_DELETE))
	    goto skip;
#endif

	if (rt_head->imported &&
		(code == NULL || as_regexp_exec (code, NULL) > 0)) {

	    t = now - rt_head->imported->time;
	    c1 = '>';
	    if (BIT_TEST (rt_head->state_bits, VRTS_SUPPRESS))
		c1 = 's';
	    if (BIT_TEST (rt_head->state_bits, VRTS_DELETE))
		c1 = 'D';

	    strcpy (tmpx, "");
	    if (BGP4_BIT_TEST (rt_head->imported->attr->attribs, 
			       PA4_TYPE_ASPATH)) {
		strcat (tmpx, aspath_toa (rt_head->imported->attr->aspath));
	        strcat (tmpx, " ");
	    }

	    if (BGP4_BIT_TEST (rt_head->imported->attr->attribs, 
			       PA4_TYPE_ORIGIN)) {
		strcat (tmpx, s_origin[rt_head->imported->attr->origin]);
	        strcat (tmpx, " ");
	    }

	    c2 = ' ';
	    rib_show_route_line (uii, c1, c2, rt_head->imported->attr->type,
				 rt_head->imported->pref, t, rt_head->prefix, 
				 rt_head->imported->attr->nexthop,
/*   (rt_head->imported->attr->link_local)? 
      rt_head->imported->attr->link_local: rt_head->imported->attr->nexthop, */
				 rt_head->imported->attr->gateway, tmpx);
	}

	LL_Iterate (rt_head->ll_routes, route) {

	    assert (rt_head->active);
	    c1 = (rt_head->imported) ? '*' : '>';
	    c1 = (route == rt_head->active) ? c1 : ' ';
	    if (BIT_TEST (rt_head->state_bits, VRTS_SUPPRESS))
		c1 = 's';
	    if (BIT_TEST (rt_head->state_bits, VRTS_DELETE))
		c1 = 'D';
	    t = now - route->time;

	    if (BGP4_BIT_TEST (route->attr->attribs, PA4_TYPE_ASPATH)) {
		if (code && as_regexp_exec (code, route->attr->aspath) < 0)
		    continue;
		sprintf (tmpx, "%s ", aspath_toa (route->attr->aspath));
	    }
	    else {
		if (code && as_regexp_exec (code, NULL) < 0)
		    continue;
	    }

	    if (BGP4_BIT_TEST (route->attr->attribs, PA4_TYPE_ORIGIN))
		sprintf (tmpx + strlen (tmpx), "%s", 
			 s_origin[route->attr->origin]);

	    c2 = (route->attr->gateway->AS == BGP->my_as)? 'i': ' ';
	    c2 = (route->aggregated)? 'a': c2;
	    rib_show_route_line (uii, c1, c2, route->attr->type,
				 route->pref, t,
				 rt_head->prefix,
    				 route->attr->nexthop,
/* (route->attr->link_local)? route->attr->link_local: route->attr->nexthop, */
				 route->attr->gateway, tmpx);

	}
}
#ifdef notdef
      skip:
#endif
    }
    VIEW_RADIX_WALK_END;
}


/* show_bgp_routing table
 * dump BGP routes to socket. usually called by UII
 */
int 
show_bgp_rt_regexp (uii_connection_t * uii, char *expr)
{
    as_regexp_code_t *code = NULL;
    int pos = 0;

    if (expr && (code = as_regexp_comp (expr, &pos)) == NULL) {
    	uii_send_data (uii, "%s\n", expr);
    	uii_send_data (uii, "%*c\n", pos, '^');
	Delete (expr);
	return (-1);
    }
    Delete (expr);

    pthread_mutex_lock (&BGP->mutex_lock);
    view_open (BGP->views[0]);
    bgp_dump_view (uii, BGP->views[0], code, NULL);
    view_close (BGP->views[0]);
    pthread_mutex_unlock (&BGP->mutex_lock);
#ifdef HAVE_IPV6
    uii_send_data (uii, "\n");
    pthread_mutex_lock (&BGP->mutex_lock);
    view_open (BGP->views[1]);
    bgp_dump_view (uii, BGP->views[1], code, NULL);
    view_close (BGP->views[1]);
    pthread_mutex_unlock (&BGP->mutex_lock);
#endif /* HAVE_IPV6 */
    if (code) Delete (code);
    return (1);
}


/* show_bgp_routing table
 * dump BGP routes to socket. usually called by UII
 */
int 
show_bgp_routing_table (uii_connection_t * uii)
{
    return (show_bgp_rt_regexp (uii, NULL));
}


#ifdef notdef
/* bgp_get_attr
 * attribute are stored in global hash. This 
 * 1) fetches attr if exists 
 * 2) or creates new attr
 */
bgp_attr_t *
bgp_get_attr (bgp_attr_t * attr)
{
    bgp_attr_t *tmpx;

    tmpx = HASH_Lookup (BGP->attr_hash, attr);

    if (tmpx == NULL) {
	HASH_Insert (BGP->attr_hash, bgp_ref_attr (attr));
	tmpx = attr;
    }

    return (tmpx);
}
#endif


int
show_bgp_neighbors_errors (uii_connection_t * uii, char *peer_or_star)
{
    bgp_peer_t *peer;

/* XXX needs lock/unlock */
    if (strcmp (peer_or_star, "*") != 0) {
	prefix_t *prefix = ascii2prefix (0, peer_or_star);
	/* XXX this happens due to a bug */
	if (prefix == NULL) {
	    if ((peer = Find_BGP_Peer_ByID (peer_or_star)) != NULL) {
		print_error_list (uii, peer->trace);
	    }
	    else {
                config_notice (TR_ERROR, uii, "Peer %s does not exist\n",
			       peer_or_star);
	        Delete (peer_or_star);
	        return (-1);
	    }
	}
        else if ((peer = Find_BGP_Peer (prefix, NULL, NULL)) == NULL) {
            config_notice (TR_ERROR, uii, "Peer %s does not exist\n",
		prefix_toa (prefix));
            Deref_Prefix (prefix);
	    Delete (peer_or_star);
            return (-1);
        }
	else {
	    print_error_list (uii, peer->trace);
            Deref_Prefix (prefix);
	}
    }
    else {
        LL_Iterate (BGP->ll_bgp_peers, peer) {
	    print_error_list (uii, peer->trace);
	}
    }
    Delete (peer_or_star);
    return (1);
}


int
show_bgp_neighbors_routes (uii_connection_t * uii, char *peer_or_star)
{
    bgp_peer_t *peer;

/* XXX needs lock/unlock */
    if (strcmp (peer_or_star, "*") != 0) {
	prefix_t *prefix = ascii2prefix (0, peer_or_star);
	/* XXX this happens due to a bug */
	if (prefix == NULL) {
	    if ((peer = Find_BGP_Peer_ByID (peer_or_star)) != NULL) {
		print_error_list (uii, peer->trace);
	    }
	    else {
                config_notice (TR_ERROR, uii, "Peer %s does not exist\n",
			       peer_or_star);
	        Delete (peer_or_star);
	        return (-1);
	    }
	}
        else if ((peer = Find_BGP_Peer (prefix, NULL, NULL)) == NULL) {
            config_notice (TR_ERROR, uii, "Peer %s does not exist\n",
			   prefix_toa (prefix));
            Deref_Prefix (prefix);
	    Delete (peer_or_star);
            return (-1);
        }
	else {
	    bgp_dump_view (uii, NULL, NULL, peer);
            Deref_Prefix (prefix);
	}
    }
    else {
	int first = 1;
        LL_Iterate (BGP->ll_bgp_peers, peer) {
            if (first) first = 0;
	    else uii_send_data (uii, "\n");
	    bgp_dump_view (uii, NULL, NULL, peer);
	}
    }
    Delete (peer_or_star);
    return (1);
}


int
bgp_check_attr (bgp_peer_t * peer, bgp_attr_t * attr, int as)
{
    int peer_as;

    if (attr == NULL)
	return (0);

    assert (peer);

    if (!BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP)) {
/* XXX */
/* This is not sufficient because there could be both IPv4 and IPv6 info */
	trace (TR_ERROR, peer->trace, "attribute: nexthop missing\n");
	bgp_send_notify_byte (peer, BGP_ERR_UPDATE, BGP_ERRUPD_MISSING, 
			      PA4_TYPE_NEXTHOP);
	bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (-1);
	
    }
    else {
	/* XXX should check nexthop well */
    }

    if (!BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH)) {
	trace (TR_ERROR, peer->trace, "attribute: aspath missing\n");
	bgp_send_notify_byte (peer, BGP_ERR_UPDATE, BGP_ERRUPD_MISSING, 
			      PA4_TYPE_ASPATH);
	bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (-1);
    }
    else {
        if (as != peer->gateway->AS /* eBGP */ &&
	    !BIT_TEST (peer->options, BGP_TRANSPARENT_AS)) {
	    if ((peer_as = bgp_get_home_AS (attr->aspath)) <= 0) {
		trace (TR_ERROR, peer->trace,
		       "attribute: no home for eBGP\n");
	        bgp_send_notification (peer, BGP_ERR_UPDATE, 
				      BGP_ERRUPD_ATTRLIST);
	        bgp_sm_process_event (peer, BGPEVENT_ERROR);
		return (-1);
	    }
	    else if (peer->gateway->AS != peer_as) {
		trace (TR_ERROR, peer->trace,
		       "attribute: strange home in aspath %s as %d\n",
		       aspath_toa (attr->aspath), peer->gateway->AS);
	        bgp_send_notification (peer, BGP_ERR_UPDATE, 
				      BGP_ERRUPD_ATTRLIST);
	        bgp_sm_process_event (peer, BGPEVENT_ERROR);
		return (-1);
	    }
	}
	if (as > 0 && bgp_check_aspath_loop (attr->aspath, as)) {
	    trace (TR_WARN, peer->trace,
		   "attribute: my own as %d in aspath %s\n",
		   as, aspath_toa (attr->aspath));
	    /* just discard */
	    return (-1);
	}
    }

    if (!BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGIN)) {
	trace (TR_ERROR, peer->trace, "attribute: origin missing\n");
	bgp_send_notify_byte (peer, BGP_ERR_UPDATE, BGP_ERRUPD_MISSING, 
			      PA4_TYPE_ORIGIN);
	bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (-1);
    }

    if (as == peer->gateway->AS /* iBGP */
	 && !BGP4_BIT_TEST (attr->attribs, PA4_TYPE_LOCALPREF)) {
	trace (TR_ERROR, peer->trace, "attribute: localpref missing\n");
	bgp_send_notify_byte (peer, BGP_ERR_UPDATE, BGP_ERRUPD_MISSING, 
			      PA4_TYPE_LOCALPREF);
	bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (-1);
    }
    return (1);
}


int       
check_bgp_networks (prefix_t *prefix)
{
    prefix_t *network;

    assert (BGP->ll_networks);
    LL_Iterate (BGP->ll_networks, network) {
        if (a_include_b (network, prefix))
            return (1);
    }
    return (0);
}
