/*
 * $Id: view.c,v 1.28 1998/09/01 15:16:56 masaki Exp $
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <time.h>
#include <ctype.h>

#include <mrt.h>
#include <config_file.h>
#include <stack.h>
#include <rib.h>
#include <bgp.h>
#include <interface.h>
#ifdef HAVE_IPV6
#include <api6.h>
#endif /* HAVE_IPV6 */

static aggregate_t *
view_check_aggregate (view_t * view, prefix_t *prefix, int withdraw);

void static
bgp_rt_update_call (int cmd, prefix_t * prefix, bgp_attr_t * attr,
		    u_long pref, u_long state_bits)
{
    u_long option = 0;

    assert (prefix);
    assert (attr);
    if (BGP->rt_update_call_fn) {
	generic_attr_t *gattr;
        gattr = New_Generic_Attr (attr->type);
        gattr->gateway = attr->gateway;
	if (attr->nexthop2) {
            gattr->nexthop = Ref_Prefix (attr->nexthop2);
	}
#ifdef HAVE_IPV6
	else if (attr->link_local) {
            gattr->nexthop = Ref_Prefix (attr->link_local);
	}
#endif /* HAVE_IPV6 */
	else {
	    assert (attr->nexthop);
            gattr->nexthop = Ref_Prefix (attr->nexthop);
	}
	if (BIT_TEST (state_bits, VRTS_SUPPRESS))
	    BIT_SET (option, MRT_RTOPT_SUPPRESS);
	/* a route learned from iBGP will not be announced by IGP */
        if (attr->gateway->AS == BGP->my_as)
	    BIT_SET (option, MRT_RTOPT_SUPPRESS);
	BGP->rt_update_call_fn (cmd, prefix, gattr, pref, option);
	Deref_Generic_Attr (gattr);
    }
}


static bgp_route_t *
New_Bgp_Route (bgp_route_head_t * bgp_route_head, bgp_attr_t * attr, int pref)
{
    bgp_route_t *route = New (bgp_route_t);

    route->head = bgp_route_head;
    route->pref = pref;
    route->attr = bgp_ref_attr (attr);
    time (&route->time);
    BGP->bgp_num_active_route_node++;
    trace (TR_TRACE, BGP->trace, "New Route: %s nh %s proto %s\n",
	   prefix_toax (bgp_route_head->prefix),
	   prefix_toa (attr->nexthop), proto2string (attr->type));
    if (attr->type == PROTO_BGP)
        LL_Add (bgp_route_head->ll_routes, route);
    else
        bgp_route_head->imported = route;
    return (route);
}


static void 
Delete_Bgp_Route (bgp_route_t * rt_node)
{
    trace (TR_TRACE, BGP->trace, "Delete Route: %s nh %s proto %s\n",
	   prefix_toax (rt_node->head->prefix),
	   prefix_toa (rt_node->attr->nexthop),
	   proto2string (rt_node->attr->type));
    bgp_deref_attr (rt_node->attr);
    Delete (rt_node);
    BGP->bgp_num_active_route_node--;
}


static bgp_route_head_t *
New_Bgp_Route_Head (prefix_t * prefix)
{
    bgp_route_head_t *route_head;

    assert (prefix);
    route_head = New (bgp_route_head_t);
    route_head->prefix = Ref_Prefix (prefix);
    route_head->ll_routes = LL_Create (LL_DestroyFunction, Delete_Bgp_Route, 0);
    route_head->active = NULL;
    route_head->imported = NULL;
    BGP->bgp_num_active_route_head++;
    trace (TR_TRACE, BGP->trace, "Add Route Head: %s\n", prefix_toax (prefix));
    return (route_head);
}


static void 
Delete_Bgp_Route_Head (bgp_route_head_t * head)
{
    /* This has done in view_delete_bgp_route */
    trace (TR_TRACE, BGP->trace, "Delete Route Head: %s\n",
	   prefix_toax (head->prefix));
    if (head->imported)
	Delete_Bgp_Route (head->imported);
    LL_Destroy (head->ll_routes);
    Deref_Prefix (head->prefix);
    Delete (head);
    BGP->bgp_num_active_route_head--;
}


static update_bucket_t *
New_Update_Bucket ()
{
    update_bucket_t *backet = New (update_bucket_t);

    /* NOTE: update bucket memory belongs to view */
    backet->ll_prefix = LL_Create (LL_DestroyFunction, Deref_Prefix, 0);
    return (backet);
}


static void 
delete_update_bucket (update_bucket_t * update_bucket)
{
    LL_Destroy (update_bucket->ll_prefix);
    bgp_deref_attr (update_bucket->attr);
    Delete (update_bucket);
}


static bgp_route_head_t *
view_find_route_head (view_t * view, prefix_t * prefix)
{
    radix_node_t *radix_node;

    radix_node = radix_search_exact (view->radix_tree, prefix);
    if (radix_node) {
	assert (radix_node->data);
	return (RADIX_DATA_GET (radix_node, bgp_route_head_t));
    }
    return (NULL);
}


static void
view_remove_route_head (view_t * view, bgp_route_head_t * route_head)
{
    assert (route_head->radix_node);
    assert (RADIX_DATA_GET (route_head->radix_node, bgp_route_head_t)
	    == route_head);
    BIT_SET (route_head->state_bits, VRTS_DELETE);
    radix_remove (view->radix_tree, route_head->radix_node);
    view->num_active_routes--;
}


static int
bgp_compare_routes (bgp_route_t * a, bgp_route_t * b)
{

    int a1, a2;

    /* these values may be u_long, so I can't subtract them to compare */

    /* smaller pref (administrative distance) wins */
    if (a->pref < b->pref)
	return (-1);
    if (a->pref > b->pref)
	return (1);

    /* tie breaking */

#ifdef notdef
    /* smaller pref2 wins */
    if (a->pref2 < b->pref2)
	return (-1);
    if (a->pref2 > b->pref2)
	return (1);
#endif

    /* check local preference */
    if (BGP4_BIT_TEST (a->attr->attribs, PA4_TYPE_LOCALPREF) &&
        BGP4_BIT_TEST (b->attr->attribs, PA4_TYPE_LOCALPREF)) {

        /* higer local_pref wins */
        if (a->attr->local_pref < b->attr->local_pref)
	    return (1);
        if (a->attr->local_pref > b->attr->local_pref)
	    return (-1);
    }

    /* shorter aspath wins */
    a1 = aspath_length (a->attr->aspath, 0);
    a2 = aspath_length (b->attr->aspath, 0);
    if (a1 < a2)
	return (-1);
    if (a1 > a2)
	return (1);

    /* lowest origin code wins */
    if (a->attr->origin < b->attr->origin)
	return (-1);
    if (a->attr->origin > b->attr->origin)
	return (1);

    /* XXX should be configurable */
    /* lower metric wins */
    if (a->attr->multiexit < b->attr->multiexit)
        return (-1);
    if (a->attr->multiexit > b->attr->multiexit)
	return (1);

    /* shorter aspath wins */

    if (a->attr->gateway && b->attr->gateway) {
        /* smaller routerid wins */
        if (ntohl (a->attr->gateway->routerid) < 
		ntohl (b->attr->gateway->routerid))
	    return (-1);
        if (ntohl (a->attr->gateway->routerid) > 
		ntohl (b->attr->gateway->routerid))
	    return (1);
    }

    return (0);
}


static bgp_route_t *
bgp_find_best_route (bgp_route_head_t * bgp_route_head)
{

    bgp_route_t *best = NULL, *route;

    /* find new active route */
    LL_Iterate (bgp_route_head->ll_routes, route) {

	/* no best */
	if (best == NULL) {
	    best = route;
	    continue;
	}

	if (bgp_compare_routes (route, best) < 0) {
	    best = route;
	    continue;
	}

    }
    return (best);
}


static int
view_add_ll_list (view_t *view, bgp_route_head_t *bgp_route_head, int withdraw)
{
    bgp_route_head_t *rh;
    LINKED_LIST *ll;

    assert (view);
    assert (bgp_route_head);

    ll = (withdraw)? view->ll_with_routes: view->ll_ann_routes;

    LL_Iterate (ll, rh) {
	if (rh == bgp_route_head) {
	    /* this is warning for a while to check out this */
	    trace (TR_WARN, view->trace,
		   "Duplicate %s prefix: %s\n",
		   (withdraw)? "withdraw": "announce",
		   prefix_toax (bgp_route_head->prefix));
	    /* it may be better to remove the previous one */
	    LL_RemoveFn (ll, rh, NULL);
	    return (0);
	}
    }
    /* we could check withdraw list because implicit withdraw is OK */
    LL_Add (ll, bgp_route_head);
    return (1);
}


static int
view_add_change_list (view_t * view, bgp_route_head_t * bgp_route_head,
		      int withdraw)
{
    aggregate_t *agg;

    agg = view_check_aggregate (view, bgp_route_head->prefix, withdraw);
    if (agg) {
        if (BIT_TEST (agg->option, BGP_AGGOPT_SUMMARY_ONLY)) {
	    BIT_SET (bgp_route_head->state_bits, VRTS_SUPPRESS);
	    trace (TR_TRACE, view->trace,
	           "Change: %s (but suppressed by %s)\n",
	           prefix_toax (bgp_route_head->prefix),
	           prefix_toax (agg->prefix));
	    return (-1);
        }
    }

    /* add the route head to the change list */
    view_add_ll_list (view, bgp_route_head, withdraw);
    trace (TR_TRACE, view->trace, "Change: %s (%s)\n",
	   prefix_toax (bgp_route_head->prefix),
	   (withdraw) ? "withdraw" : "announce");

    return (1);
}


static bgp_route_t *
view_add_bgp_route (view_t * view, bgp_route_head_t * bgp_route_head,
		    bgp_attr_t * attr, u_long pref)
{
    bgp_route_t *bgp_route;

    LL_Iterate (bgp_route_head->ll_routes, bgp_route) {
	if (bgp_route->attr->gateway == attr->gateway)
	    assert (0);
    }

    /* first time we've heard from this gateway */
    bgp_route = New_Bgp_Route (bgp_route_head, attr, pref);

    /* is this new route better than the active route ? */
    if (bgp_route_head->active &&
	bgp_compare_routes (bgp_route, bgp_route_head->active) >= 0)
	return (bgp_route);

    if (view->rib)
	bgp_rt_update_call (RIB_UPD, bgp_route_head->prefix,
			    bgp_route->attr, bgp_route->pref,
			    bgp_route_head->state_bits);
    bgp_route_head->active = bgp_route;


    /*
     * If this becomes the best in the RIB, 
     * the imported will be updated from the RIB.
     */

    /* decision by this proto instead of by rib */
    /* this is needed to suppress contributed routes by the aggregate addr */
    if (bgp_route_head->imported &&
	bgp_route_head->imported->pref >= bgp_route->pref) {
	Delete_Bgp_Route (bgp_route_head->imported);
	bgp_route_head->imported = NULL;
    }
    if (bgp_route_head->imported == NULL) {
	/* add the route head to the change list */
	view_add_change_list (view, bgp_route_head, 0);
    }

    return (bgp_route);
}


/* 
 * delete a route from view. If active route, set change list
 * and modify route_head
 */
static int
view_delete_bgp_route (view_t * view, bgp_route_t * bgp_route)
{
    bgp_route_head_t *bgp_route_head;
    bgp_route_t *best;

    best = NULL;
    bgp_route_head = bgp_route->head;

    /* remove from list in any case without destroying it */
    LL_RemoveFn (bgp_route_head->ll_routes, bgp_route, NULL);

    /* we are NOT the active route - just delete */
    if (bgp_route_head->active != bgp_route) {
	Delete_Bgp_Route (bgp_route);
	return (1);
    }

    /* we were the active route */

    /* find new active route */
    best = bgp_find_best_route (bgp_route_head);

    /* prefix withdrawn with nothing to replace it */
    if (best == NULL) {

	if (bgp_route->attr->type == PROTO_BGP && view->rib)
	    bgp_rt_update_call (RIB_DEL, bgp_route_head->prefix,
				bgp_route->attr, 0, 0);

	if (bgp_route_head->imported) {
	    Delete_Bgp_Route (bgp_route);
	    bgp_route_head->active = NULL;
	    return (0);
	}

	/* remove route_head first */
	view_remove_route_head (view, bgp_route_head);
	/* the route pointed by active will be freed later */
	view_add_change_list (view, bgp_route_head, 1);
	return (1);
    }

    /* the new active found */
    if (view->rib) {
	bgp_rt_update_call (RIB_UPD, bgp_route_head->prefix,
			best->attr, best->pref, bgp_route_head->state_bits);
    }

    Delete_Bgp_Route (bgp_route);
    /* set to new best */
    bgp_route_head->active = best;

    if (bgp_route_head->imported == NULL) {
	view_add_change_list (view, bgp_route_head, 0);
    }

    return (1);
}


/*
 * this will be called from rib, that is, they are called from
 * other threads. Schedule it.
 */

static void
bgp_import (prefix_t * prefix, generic_attr_t * new,
	    generic_attr_t * old, u_int pref)
{
    bgp_attr_t *bgp_new = NULL, *bgp_old = NULL;
    int i;
    view_t *view;

    assert (new == NULL || new->type != PROTO_BGP);
    assert (old == NULL || old->type != PROTO_BGP);

    if (old) {
	trace (TR_TRACE, BGP->trace, "Imported Delete: %s nh %s proto %s\n",
	       prefix_toax (prefix),
	       prefix_toa (old->nexthop),
	       proto2string (old->type));
	bgp_old = bgp_new_attr (old->type);
	bgp_old->gateway = old->gateway;
	if (old->nexthop) {
	    bgp_old->nexthop = Ref_Prefix (old->nexthop);
            BGP4_BIT_SET (bgp_old->attribs, PA4_TYPE_NEXTHOP);
	}
    }
    if (new) {
	trace (TR_TRACE, BGP->trace, "Imported Add: %s nh %s proto %s\n",
	       prefix_toax (prefix),
	       prefix_toa (new->nexthop),
	       proto2string (new->type));
	bgp_new = bgp_new_attr (new->type);
	bgp_new->gateway = new->gateway;
	if (new->nexthop) {
	    bgp_new->nexthop = Ref_Prefix (new->nexthop);
            BGP4_BIT_SET (bgp_new->attribs, PA4_TYPE_NEXTHOP);
	}

	assert (new->type != PROTO_BGP);
	/* XXX need to check if cisco makes even static routes igp */
	if ((new->type == PROTO_CONNECTED || new->type == PROTO_STATIC) 
		&& check_bgp_networks (prefix))
            bgp_new->origin = 0;
	else if (new->type == PROTO_RIPNG || new->type == PROTO_RIP ||
	         new->type == PROTO_OSPF)
            bgp_new->origin = 0;
        else
            bgp_new->origin = 2;
        BGP4_BIT_SET (bgp_new->attribs, PA4_TYPE_ORIGIN);

        if (!BGP4_BIT_TEST (bgp_new->attribs, PA4_TYPE_ASPATH)) {
	    bgp_new->aspath = NULL;
            BGP4_BIT_SET (bgp_new->attribs, PA4_TYPE_ASPATH);
	}
    }

    if (bgp_new == NULL && bgp_old == NULL)
	return;

    for (i = 0; i < MAX_BGP_VIEWS; i++) {

	if ((view = BGP->views[i]) == NULL ||
	    view->family != prefix->family)
	    continue;

	view_open (view);

	if (old) {
	    bgp_del_route (view, prefix, bgp_old);
	}

	if (new) {
	    bgp_add_route (view, prefix, bgp_new, pref);
	}

	bgp_process_changes (view);
	view_close (view);
    }

    Deref_Prefix (prefix);
    if (new) {
	Deref_Generic_Attr (new);
	bgp_deref_attr (bgp_new);
    }
    if (old) {
	Deref_Generic_Attr (old);
	bgp_deref_attr (bgp_old);
    }
}


static LINKED_LIST *
build_update_buckets (LINKED_LIST * ll_routes)
{
    LINKED_LIST *ll_update_buckets;
    update_bucket_t *update_bucket;
    bgp_route_t *route;
    int found;

    ll_update_buckets = LL_Create (LL_DestroyFunction, delete_update_bucket, 0);

    /* replace this with a hash, or something a bit more reasonable */
    LL_Iterate (ll_routes, route) {
	found = 0;
	LL_Iterate (ll_update_buckets, update_bucket) {
	    if (update_bucket->attr == route->attr) {
		LL_Add (update_bucket->ll_prefix,
			Ref_Prefix (route->head->prefix));
		found = 1;
		break;
	    }
	}

	if (found == 0) {
	    update_bucket = New_Update_Bucket ();
	    update_bucket->attr = bgp_ref_attr (route->attr);

	    LL_Add (ll_update_buckets, update_bucket);
	    LL_Add (update_bucket->ll_prefix,
		    Ref_Prefix (route->head->prefix));
	}
    }

    return (ll_update_buckets);
}


/*
 * get "head" of routing table entry for a given prefix. Create a head
 * and radix-tree entry if one does not aleady exist
 */
static bgp_route_head_t *
view_get_route_head (view_t * view, prefix_t * prefix)
{
    bgp_route_head_t *bgp_route_head;
    radix_node_t *radix_node;

    assert (view);
    assert (prefix);
    radix_node = radix_search_exact (view->radix_tree, prefix);

    if (radix_node) {
	assert (radix_node->data);
	bgp_route_head = RADIX_DATA_GET (radix_node, bgp_route_head_t);
    }
    else {
	bgp_route_head = New_Bgp_Route_Head (prefix);
	radix_node = radix_lookup (view->radix_tree, prefix);
	assert (radix_node->data == NULL);
	RADIX_DATA_SET (radix_node, bgp_route_head);
	bgp_route_head->radix_node = radix_node;
	view->num_active_routes++;
    }
    return (bgp_route_head);
}


static int
view_announce_update (bgp_peer_t * peer, u_long pdu_len, u_char * start_pdu)
{
    return (schedule_event2 ("bgp_send_update", peer->schedule, 
			     (void_fn_t) bgp_send_update, 
			     3, peer, pdu_len, start_pdu));
}


void
view_set_nexthop_self (bgp_peer_t *peer, bgp_attr_t *attr)
{

    if (attr->nexthop) {
	Deref_Prefix (attr->nexthop);
	attr->nexthop = NULL;
        BGP4_BIT_RESET (attr->attribs, PA4_TYPE_NEXTHOP);
    }

    if ((attr->nexthop = Ref_Prefix (peer->local_addr)) != NULL) {
        BGP4_BIT_SET (attr->attribs, PA4_TYPE_NEXTHOP);
#ifdef HAVE_IPV6
        if (attr->nexthop->family == AF_INET6) {
#define CISCO_BGP4MP
#ifdef CISCO_BGP4MP
	    /* BGP4+ doesn't require ipv4 nexthop for MP NLRI,
	       but cisco expects */
            static prefix_t *my_id_prefix = NULL;

            if (my_id_prefix == NULL)
                my_id_prefix = New_Prefix (AF_INET, &BGP->my_id, 32);

            if (attr->nexthop4)
	        Deref_Prefix (attr->nexthop4);
            attr->nexthop4 = Ref_Prefix (my_id_prefix);
#endif /* CISCO_BGP4MP */

            if (attr->link_local)
	        Deref_Prefix (attr->link_local);
            if (peer->gateway && peer->gateway->interface &&
	        peer->gateway->interface->link_local)
	        attr->link_local = 
	            Ref_Prefix (peer->gateway->interface->link_local->prefix);
            else
	        attr->link_local = NULL;
	}
#endif /* HAVE_IPV6 */
    }
}


/* 
 * here is where we actually build update packets and 
 * schedule them to be sent off. NOTE: It is is the responsibility of
 * bgp peer to delete buffer memory!
 */
/* 
 * modify attr to announce neighbors, create packets, and schedule them
 */
static int 
view_announce_peer (bgp_peer_t * peer, LINKED_LIST * ll_with_prefixes,
		    LINKED_LIST * ll_ann_prefixes, bgp_attr_t *attr)
{
    int as = BGP->my_as;

    assert (peer);

    /* attr may be NULL in case of withdraw */
    if (attr) {

	attr = bgp_copy_attr (attr);

	if (BIT_TEST (peer->options, BGP_TRANSPARENT_NEXTHOP)) {
	    /* OK, nothing to do */
	}
	else if (attr->type == PROTO_CONNECTED) {
	   /*
	    * always set the next hop with a local address
	    */
	    view_set_nexthop_self (peer, attr);
	}
	else if (peer->gateway->AS == BGP->my_as) { /* iBGP */
	    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP) &&
	       (prefix_is_unspecified (attr->nexthop) ||
	        prefix_is_loopback (attr->nexthop))) {
	        view_set_nexthop_self (peer, attr);
	    }
	}
	else if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP)) {
	    if (prefix_is_unspecified (attr->nexthop) ||
	        prefix_is_loopback (attr->nexthop)) {
	        view_set_nexthop_self (peer, attr);
	    }
	    else if (!(peer->gateway && peer->gateway->interface &&
	           is_prefix_on (attr->nexthop, peer->gateway->interface))) {
	        view_set_nexthop_self (peer, attr);
	    }
	    else if (BIT_TEST (peer->options, BGP_EBGP_MULTIHOP)) {
	        view_set_nexthop_self (peer, attr);
#ifdef HAVE_IPV6
	        if (attr->link_local) {
		    Deref_Prefix (attr->link_local);
		    attr->link_local = NULL;
		}
#endif /* HAVE_IPV6 */
	    }

	}
	else {
	    /* replace nexthop with local address */
	    view_set_nexthop_self (peer, attr);
	}

	if (peer->route_map_out > 0) {
	    bgp_trace_attr2 (attr, 0, peer->trace);
            attr = apply_route_map (peer->route_map_out, attr);
	    trace (TR_TRACE, peer->trace,
		   "route-map %d applied for output\n", 
		   peer->route_map_out);
	}

        if (BIT_TEST (peer->options, BGP_TRANSPARENT_AS)) {
            as = 0;
	}
#ifdef notdef
	else {
	    if (!BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH)) {
	        attr->aspath = NULL;
	        BGP4_BIT_SET (attr->attribs, PA4_TYPE_ASPATH);
	    }
	}
#endif

        /* iBGP */
	if (peer->gateway->AS == BGP->my_as) {
	    as = 0;
	    if (!BGP4_BIT_TEST (attr->attribs, PA4_TYPE_LOCALPREF)) {
	        attr->local_pref = BGP->default_local_pref;
	        BGP4_BIT_SET (attr->attribs, PA4_TYPE_LOCALPREF);
	    }
	}
	/* eBGP */
	else {
	    BGP4_BIT_RESET (attr->attribs, PA4_TYPE_LOCALPREF);
	}

#ifdef notdef
	/* there must be an origin info */
	if (!BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGIN)) {
	    attr->origin = 0;
	    BGP4_BIT_SET (attr->attribs, PA4_TYPE_ORIGIN);
	}
#endif
    }

    bgp_create_pdu (ll_with_prefixes, ll_ann_prefixes, attr, as,
			    peer, view_announce_update, 
			    NULL, 0, 0);
    if (attr)
        bgp_deref_attr (attr);
    return (0);
}


static bgp_route_t *
bgp_implicit_withdraw (view_t * view, bgp_route_t * bgp_route,
		       bgp_attr_t * attr, u_long pref)
{
    bgp_route_head_t *bgp_route_head;
    bgp_route_t *best;

    best = NULL;
    bgp_route_head = bgp_route->head;

    if (bgp_route->pref == pref &&
	    bgp_compare_attr (bgp_route->attr, attr) > 0) {
        trace (TR_TRACE, view->trace,
	       "Implicit withdraw but same: %s (ignore)\n",
	       prefix_toax (bgp_route_head->prefix));
	/* Do we need to update the time received? */
        return (bgp_route);
    }

    /* remove old route from list in any case and destroy it */
    LL_Remove (bgp_route_head->ll_routes, bgp_route);

    /* if we were active route, wipe clean active route */
    if (bgp_route_head->active == bgp_route) {
	bgp_route_head->active = NULL;
    }

    /* and add new route */
    bgp_route = New_Bgp_Route (bgp_route_head, attr, pref);

    /* find new active route */
    best = bgp_find_best_route (bgp_route_head);

    assert (best);

    if (best && bgp_route_head->active == best) {
	return (bgp_route);
    }

    /* set to new best */
    bgp_route_head->active = best;
    if (view->rib)
	bgp_rt_update_call (RIB_UPD, bgp_route_head->prefix,
			best->attr, best->pref, bgp_route_head->state_bits);

    if (bgp_route_head->imported == NULL ||
    /* decision by this proto instead of by rib */
    /* this is needed to suppress contributed routes by the aggregate addr */
	bgp_route_head->imported->pref >= best->pref) {
	if (bgp_route_head->imported) {
	    Delete_Bgp_Route (bgp_route_head->imported);
	    bgp_route_head->imported = NULL;
	}
	/* add the route head to the change list */
	view_add_change_list (view, bgp_route_head, 0);
    }

    return (bgp_route);
}


/*
 * tentative for a bgp policy routing...
 */
static int
bgp_policy (prefix_t * prefix, bgp_attr_t * attr, bgp_peer_t * peer, int inout)
{
    int num;

    assert (peer);
    if (inout == 0 /* announce */|| inout == 2 /* withdraw */) {
	/* input policy processing */

	/* verify the prefix */
#ifdef HAVE_IPV6
	if (prefix->family == AF_INET) {
#endif /* HAVE_IPV6 */

	    if (prefix->bitlen > 32) {
		trace (TR_PACKET, peer->trace,
		       "  x %s (invalid prefix length)\n",
		       prefix_toax (prefix));
		return (0);
	    }
#ifdef HAVE_IPV6
	}
	else {
	    assert (prefix->family == AF_INET6);

	    if (prefix->bitlen > 128) {
		trace (TR_PACKET, peer->trace,
		       "  x %s (invalid prefix length)\n",
		       prefix_toax (prefix));
		return (0);
	    }
	}
#endif /* HAVE_IPV6 */

#ifdef notdef
/* bgpsim doesn't put a gateway */
	if (attr->gateway == NULL)
	    return (1);
#endif

#ifdef notdef
/* XXX it might be too strict */
	if (peer->local_addr->family != prefix->family) {
	    trace (TR_PACKET, peer->trace,
		   "  x %s (family %d on %d)\n",
		   prefix_toax (prefix),
		   prefix->family, peer->local_addr->family);
	    return (0);
	}
#endif

	/* check distribute-list for in */
	if ((num = peer->dlist_in) > 0) {
	    if (apply_access_list (num, prefix) == 0) {
		trace (TR_PACKET, peer->trace,
		       "  x %s (d-list %d)\n",
		       prefix_toax (prefix), num);
		return (0);
	    }
	}

	/* check filter-list for in */
	/* apply for announce only */
	if (inout == 0 && (num = peer->flist_in) > 0) {
	    if (!BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH) ||
		apply_as_access_list (num, attr->aspath) == 0) {
		trace (TR_PACKET, peer->trace,
		       "  x %s (f-list %d)\n",
		       prefix_toax (prefix), num);
		return (0);
	    }
	}
    }
    else {
	/* output policy processing */

#ifdef notdef
/* bgpsim doesn't put a gateway */
	if (attr->gateway == NULL)
	    return (1);
#endif

	/* BGPsim skips these checks */
        if (!BIT_TEST (peer->options, BGP_TRANSPARENT_AS)) {

	    /* do not announce to peer who gave us route */
	    if (attr->gateway == peer->gateway) {
		trace (TR_PACKET, peer->trace,
		       "  x %s (split horizon)\n",
		       prefix_toax (prefix));
		return (0);
	    }

	    /* do not announce to peer who once passed it */
	    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH) &&
	        bgp_check_aspath_loop (attr->aspath, 
				       peer->gateway->AS)) {
		trace (TR_PACKET, peer->trace,
		       "  x %s (peer as %d in path)\n",
		       prefix_toax (prefix), peer->gateway->AS);
		return (0);
	    }

	    /* do not announce to iBGP if larned via iBGP */
	    if (peer->gateway->AS == BGP->my_as &&
	        attr->gateway->AS == BGP->my_as) {
		trace (TR_PACKET, peer->trace,
		       "  x %s (iBGP - iBGP)\n",
		       prefix_toax (prefix));
		return (0);
	    }

	    /* do not announce connected route to peer on the interface */
	    if (attr->type == PROTO_CONNECTED && 
		    attr->gateway->interface &&
		    !BIT_TEST (peer->options, BGP_EBGP_MULTIHOP) &&
		    attr->gateway->interface == 
			peer->gateway->interface) {
		trace (TR_PACKET, peer->trace,
		       "  x %s (on the interface %s)\n",
		       prefix_toax (prefix), 
		       peer->gateway->interface->name);
		return (0);
	    }
        }

	if (attr->type != PROTO_BGP &&
	    check_bgp_networks (prefix)) {
	    /* OK */
#ifdef notdef
	    attr->origin = 0;
	    BGP4_BIT_SET (attr->attribs, PA4_TYPE_ORIGIN);
#endif
	}

	else if (attr->type != PROTO_BGP &&
	    !BIT_TEST (MRT->redist[attr->type], (1 << PROTO_BGP))) {
	    trace (TR_PACKET, peer->trace,
		   "  x %s (proto %s)\n",
		   prefix_toax (prefix),
		   proto2string (attr->type));
	    return (0);
	}

#ifdef notdef
/* XXX it might be too strict */
	if (peer->local_addr->family != prefix->family) {
	    trace (TR_PACKET, peer->trace,
		   "  x %s (family %d on %d)\n",
		   prefix_toax (prefix),
		   prefix->family, peer->local_addr->family);
	    return (0);
	}
#endif
	/* check distribute-list for out */
	if ((num = peer->dlist_out) > 0) {
	    if (apply_access_list (num, prefix) == 0) {
		trace (TR_PACKET, peer->trace,
		       "  x %s (d-list %d)\n",
		       prefix_toax (prefix), num);
		return (0);
	    }
	}

	/* check filter-list for out */
	if ((num = peer->flist_out) > 0) {
	    if (!BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH) ||
	        apply_as_access_list (num, attr->aspath) == 0) {
		trace (TR_PACKET, peer->trace,
		       "  x %s (f-list %d)\n",
		       prefix_toax (prefix), num);
		return (0);
	    }
	}
    }
    return (1);
}


/*
 * process route changes - run policy, and then build and schedule 
 * updates to peers. This is call after receipt of update packets
 */
/* public */
int 
bgp_process_changes (view_t * view)
{
    bgp_route_head_t *head;
    bgp_route_t *route = NULL;
    LINKED_LIST *ll_ann = NULL;
    LINKED_LIST *ll_with = NULL;
    LINKED_LIST *ll_buckets = NULL;
    bgp_peer_t *peer;
    update_bucket_t *bucket = NULL;;

    /* maybe check that view is open? */

    /* update each peer of view with changes */
    LL_Iterate (BGP->ll_bgp_peers, peer) {

	if (BGP->views[peer->view] != view)
	    continue;

	/* see if peer is established */
	if (peer->state != BGPSTATE_ESTABLISHED)
	    continue;

	ll_ann = LL_Create (0);
	ll_with = LL_Create (0);

	/* build withdraw list */
	LL_Iterate (view->ll_with_routes, head) {

	    /* 
	     * Withdraw 
	     */

	    if (head->active)
		route = head->active;
	    if (head->imported)
		route = head->imported;
	    assert (route);

	    /* do not give withdraw to person who gave us the route */
	    /* if not announced, then do not withdraw */
	    if (route->attr->gateway == peer->gateway ||
		!BIT_TEST (head->peer_mask, peer->bit))
		continue;

	    BIT_RESET (head->peer_mask, peer->bit); /* reset announce bit */
	    LL_Add (ll_with, head->prefix);
	}

	/* build announce list */
	LL_Iterate (view->ll_ann_routes, head) {

	    /* 
	     * Add
	     */

	    if (head->active && head->imported) {
		route = (head->active->pref <= head->imported->pref) ?
		    head->active : head->imported;
	    }
	    else if (head->imported)
		route = head->imported;
	    else if (head->active)
		route = head->active;
	    assert (route);

	    /* policy */
	    if (!bgp_policy (head->prefix, route->attr, peer, 1)) {
		/* if policy fails and route announced, withdraw the puppy */
		if (BIT_TEST (head->peer_mask, peer->bit)) {
	    	    /* reset announce bit */
		    BIT_RESET (head->peer_mask, peer->bit); 
		    LL_Add (ll_with, head->prefix);
		}
		continue;
	    }
	    BIT_SET (head->peer_mask, peer->bit);	/* set announce bit */
	    LL_Add (ll_ann, route);
	}

	/* just withdraws */
	if (LL_GetCount (ll_ann) <= 0 && LL_GetCount (ll_with) > 0) {
	    view_announce_peer (peer, ll_with, NULL, NULL);
	}
	else {

	    /* organize by attr type for easy building of packets */
	    ll_buckets = build_update_buckets (ll_ann);

	    LL_Iterate (ll_buckets, bucket) {

		/* don't announce if not sent */
		/* don't announce if we learned it from this gateway */
		if (bucket->attr->gateway == peer->gateway)
		    continue;

		view_announce_peer (peer, ll_with, bucket->ll_prefix, 
				    bucket->attr);
	    }

	    LL_Destroy (ll_buckets);
	}

	LL_Destroy (ll_ann);
	LL_Destroy (ll_with);
    }

    LL_Iterate (view->ll_with_routes, head) {
	if (BIT_TEST (head->state_bits, VRTS_DELETE)) {
	    if (head->active)
		Delete_Bgp_Route (head->active);
	    Delete_Bgp_Route_Head (head);
	}
    }
    LL_ClearFn (view->ll_with_routes, 0);
    LL_ClearFn (view->ll_ann_routes, 0);

    return (1);
}


/* public */
int
process_bgp_update (bgp_peer_t * peer, u_char * cp, int length)
{
    prefix_t *prefix;
    int i;
    int pref = BGP_PREF;
    bgp_attr_t *attr;
    LINKED_LIST *ll_announce, *ll_withdraw;

    assert (peer);

    /*
     * Okay, first of all, we dump the packet and the current route table.
     * So, it may include a packet with errors.
     */
    view_open (BGP->views[peer->view]);
    bgp_write_mrt_msg (peer, MSG_BGP_UPDATE, cp, length);
    view_close (BGP->views[peer->view]);

    if (bgp_process_update_packet (cp, length, peer) < 0) {
        bgp_send_notification (peer, peer->code, peer->subcode);
        bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (-1);
    }
    attr = peer->attr;
    ll_announce =  peer->ll_announce;
    ll_withdraw =  peer->ll_withdraw;

    if (ll_announce && attr) { /* announce */

        if (bgp_check_attr (peer, attr, BGP->my_as) < 0) {
	    /* notification was sent */
	    bgp_deref_attr (attr);
	    attr = NULL;
            if (ll_announce) {
	      /* only discard the announcements, but eventually close 
		 the connection to withdraw all heard from the peer */
		if (LL_GetCount (ll_announce) > 0) {
	            char buffer[MAXLINE], *cp = buffer;
		    LL_Iterate (ll_announce, prefix) {
        	        sprintf (cp, " %s", prefix_toax (prefix));
		        cp += strlen (cp);
		    }
	    	    trace (TR_WARN, peer->trace,
	           	    "dropping off the announce:%s\n", buffer);
		}
	        LL_Destroy (ll_announce);
		ll_announce = NULL;
	    }
        }

	/* set nexthop differntly for IBGP and EBGP -- make sure attr
	 * still exists!
	 */
	if (attr && BIT_TEST (peer->options, BGP_EBGP_MULTIHOP)) {
	    attr->nexthop2 = Ref_Prefix (peer->nexthop);
	    trace (TR_PACKET, peer->trace, "nexthop2 %s (multihop)\n", 
		    prefix_toa (attr->nexthop2));
	}
	else if (attr && (peer->gateway->AS == BGP->my_as)) {
	    attr->nexthop2 = Ref_Prefix (peer->gateway->prefix);
	    trace (TR_PACKET, peer->trace, "nexthop2 %s (iBGP)\n", 
		    prefix_toa (attr->nexthop2));
	}
    }

    if (attr == NULL) {
	attr = bgp_new_attr (PROTO_BGP);
	trace (TR_PACKET, peer->trace,
	       "Creating a pseudo attribute for the withdraw\n");
    }
    attr->gateway = peer->gateway;

    /* XXX */
    if (peer->default_pref > 0)
	pref = peer->default_pref;

    if (ll_announce || ll_withdraw) {
	/* iterate through views adding the announce/withdraws */

	for (i = 0; i < MAX_BGP_VIEWS; i++) {
	    if (BGP->views[i] == NULL)
		continue;

	    view_open (BGP->views[i]);

	    if (ll_withdraw) {

		LL_Iterate (ll_withdraw, prefix) {
		    if (BGP->views[i]->family != prefix->family)
			continue;
		    /* apply policy */
		    if (!bgp_policy (prefix, attr, peer, 2))
			continue;
		    bgp_del_route (BGP->views[i], prefix, attr);
		}
	    }

	    if (ll_announce) {
		LL_Iterate (ll_announce, prefix) {
		    if (BGP->views[i]->family != prefix->family)
			continue;
		    /* apply policy */
		    if (!bgp_policy (prefix, attr, peer, 0))
			continue;
		    if (peer->route_map_in > 0) {
		        apply_route_map (peer->route_map_in, attr);
			trace (TR_TRACE, peer->trace,
	       		       "route-map %d applied for input\n",
	       		       peer->route_map_in);
			bgp_trace_attr2 (attr, 0, peer->trace);
		    }
		    bgp_add_route (BGP->views[i], prefix, attr, pref);
		}
	    }

	    /* process change list -- send out updates to peers */
	    bgp_process_changes (BGP->views[i]);

	    view_close (BGP->views[i]);
	}
    }

    if (attr)
	bgp_deref_attr (attr);
    if (ll_withdraw)
	LL_Destroy (ll_withdraw);
    if (ll_announce)
	LL_Destroy (ll_announce);

    return (1);
}


/* public */
aggregate_t *
view_add_aggregate (view_t * view,
		    prefix_t * prefix, u_long opt)
{
    aggregate_t *agg;
    radix_node_t *radix_node;

    assert (view);
    assert (prefix);
    assert (view->family == prefix->family);

    radix_node = radix_lookup (view->agg_tree, prefix);
    assert (radix_node);

    if ((agg = RADIX_DATA_GET (radix_node, aggregate_t)) != NULL) {
	trace (TR_TRACE, view->trace,
	       "Aggregate Entry: %s already exist\n",
	       prefix_toax (prefix));
	return (agg);
    }

    agg = New (aggregate_t);
    agg->prefix = Ref_Prefix (prefix);
    agg->option = opt;
    RADIX_DATA_SET (radix_node, agg);
    agg->radix_node = radix_node;
    agg->route = NULL;

    {
	static prefix_t *my_id_prefix = NULL;
	static prefix_t *nexthop_self = NULL;
	static gateway_t *gateway_self = NULL;
#ifdef HAVE_IPV6
	static prefix_t *nexthop_self6 = NULL;
	static prefix_t *linklocal_self6 = NULL;
	static gateway_t *gateway_self6 = NULL;
#endif /* HAVE_IPV6 */

	if (my_id_prefix == NULL) {
	    my_id_prefix = New_Prefix (AF_INET, &BGP->my_id, 32);
	    nexthop_self = ascii2prefix (AF_INET, "127.0.0.1/32");
	    /* nexthop_self = ascii2prefix (AF_INET, "0.0.0.0/32"); */ /* ank */
	    gateway_self = add_gateway (nexthop_self, 0, NULL);
#ifdef HAVE_IPV6
	    nexthop_self6 = ascii2prefix (AF_INET6, "::1/128");
	    /* linklocal_self6 = ascii2prefix (AF_INET6, "fe80::1/128"); */
	    /* nexthop_self6 = ascii2prefix (AF_INET6, "::/128"); */ /* ank */
	    /* linklocal_self6 = ascii2prefix (AF_INET6, "::/128"); */ /* ank */
	    gateway_self6 = add_gateway (nexthop_self6, 0, NULL);
#endif /* HAVE_IPV6 */
	}

	agg->attr = bgp_new_attr (PROTO_BGP);
#ifdef HAVE_IPV6
	if (agg->prefix->family == AF_INET6) {
	    agg->attr->nexthop = Ref_Prefix (nexthop_self6);
	    agg->attr->link_local = Ref_Prefix (linklocal_self6);
	    agg->attr->nexthop4 = Ref_Prefix (my_id_prefix);
	    agg->attr->gateway = gateway_self6;
	}
	else
#endif /* HAVE_IPV6 */
	{
	    agg->attr->nexthop = Ref_Prefix (nexthop_self);
	    agg->attr->gateway = gateway_self;
	}
	BGP4_BIT_SET (agg->attr->attribs, PA4_TYPE_NEXTHOP);
	agg->attr->aspath = NULL;
	BGP4_BIT_SET (agg->attr->attribs, PA4_TYPE_ASPATH);
	agg->attr->origin = 3; /* Invalid. Should not be exported in this form
                                  in any case! -- alexey */;
	BGP4_BIT_SET (agg->attr->attribs, PA4_TYPE_ORIGIN);
#ifdef notdef
	if (BIT_TEST (opt, BGP_AGGOPT_SUMMARY_ONLY))
	    BGP4_BIT_SET (agg->attr->attribs, PA4_TYPE_ATOMICAGG);
	/* BIT_TEST (opt, BGP_AGGOPT_AS_SET); */
#endif
	agg->attr->aggregator.as = BGP->my_as;
	agg->attr->aggregator.prefix = Ref_Prefix (my_id_prefix);
	BGP4_BIT_SET (agg->attr->attribs, PA4_TYPE_AGGREGATOR);
    }
    trace (TR_TRACE, view->trace,
	   "AGG New Entry: %s\n",
	   prefix_toax (agg->prefix));
    bgp_trace_attr2 (agg->attr, 0, view->trace);

    return (agg);
}


/* public */
int
view_del_aggregate (view_t * view, prefix_t * prefix)
{
    aggregate_t *agg;
    radix_node_t *radix_node;

    assert (view);
    assert (prefix);
    assert (view->family == prefix->family);

    radix_node = radix_search_exact (view->agg_tree, prefix);
    if (radix_node == NULL) {
	trace (TR_TRACE, view->trace,
	       "Aggregate Entry: %s not found\n",
	       prefix_toax (prefix));
	return (-1);
    }

    agg = RADIX_DATA_GET (radix_node, aggregate_t);
    Deref_Prefix (agg->prefix);
    bgp_deref_attr (agg->attr);
    if (agg->route) {
	view_delete_bgp_route (view, agg->route);
    }
    Delete (agg);
    radix_remove (view->agg_tree, radix_node);
    return (1);
}


/* public */
void
view_print_aggregate (int viewno, void (*fn)())
{
    aggregate_t *agg;
    radix_node_t *node;
    view_t *view = BGP->views[viewno];

    if (view == NULL || view->agg_tree->head == NULL)
	return;

    RADIX_WALK (view->agg_tree->head, node) {
         agg = RADIX_DATA_GET (node, aggregate_t);
        (*fn) (agg->prefix, agg->option, viewno);
   } RADIX_WALK_END;
}


#ifdef notdef
static int
prefix_compare_inclusive (prefix_t * prefix, prefix_t * it)
{
    if (prefix->family != it->family)
	return (0);
    if (prefix->bitlen >= it->bitlen)	/* should not the same */
	return (0);

    return (comp_with_mask (prefix_tochar (prefix),
			    prefix_tochar (it),
			    prefix->bitlen));
}
#endif


static int
origin_merge (o1, o2)
{
    int origin = 0;

    if (o1 >= 2 || o2 >= 2 /* incomplete */ )
	origin = 2;
    else if (o1 == 1 || o2 == 1 /* egp */ )
	origin = 1;
    return (origin);
}


/* pseudo code Alexey (kuznet@ms2.inr.ac.ru) provide me:

> * If a route, contributing to an aggregate changes, added, removed, then:
>
>   * Find all routes, contributing to aggregate.
>   * Get the longest common AS_SEQ. Denote it AS_SEQ_HEAD.
>   * Tails of all aspathes are gathered to one AS_SET: AS_SEQ_TAIL.
>   * If AS_SEQ_TAIL is empty, goto CHECK_AAG
>   * If "as-set" is not specified, drop AS_SEQ_TAIL, set origin
>     to unspecified and goto CHECK_AAG
>
> CHECK_AAG:
>   * If atomic-aggregate is set at least on one route then
>       set atomic-agg on aggregate and RETURN.
>     * Now no routes have atomic-aggregate attr *
>   * If AS_SEQ_TAIL was empty or as-set was specified; RETURN
>   * If some routes are suppressed (by summary-only or by suppress-map),
>     and a suppresses route has AS in its aspath not present
>     in aspath of aggregate, set atomic-aggregate on aggregate.
>
> One more note: atomic aggregate should not be referenced anywhere outside
> this piece of pseudocode, because mrt does not make deaggregation.

*/


/* merge attr into attn, then return attn */
static bgp_attr_t *
view_aggregate_attr_merge (bgp_attr_t *attn, bgp_attr_t *attr, 
			   aggregate_t *agg, aspath_t *tail)
{
	    assert (attn);
	    assert (attr);
#ifdef notdef
		if (attn->gateway != attr->gateway)
		    attn->gateway = NULL;
#endif

		if (BGP4_BIT_TEST (attn->attribs, PA4_TYPE_NEXTHOP) &&
		    BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP) &&
		    prefix_compare (attn->nexthop, attr->nexthop)) {
			/* OK */
		}
		else {
		    if (BGP4_BIT_TEST (attn->attribs, PA4_TYPE_NEXTHOP)) {
		        Deref_Prefix (attn->nexthop);
		        attn->nexthop = NULL;
		        BGP4_BIT_RESET (attn->attribs, PA4_TYPE_NEXTHOP);
		    }
		}

		if (BGP4_BIT_TEST (attn->attribs, PA4_TYPE_ORIGIN) ||
		    BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGIN)) {
		    attn->origin = origin_merge (attn->origin, attr->origin);
		    BGP4_BIT_SET (attn->attribs, PA4_TYPE_ORIGIN);
		}

		if (BGP4_BIT_TEST (attn->attribs, PA4_TYPE_ASPATH) ||
		    BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH)) {
                    attn->aspath = aspath_merge (attn->aspath, attr->aspath, tail);
		    BGP4_BIT_SET (attn->attribs, PA4_TYPE_ASPATH);
		}

		if (BGP4_BIT_TEST (attn->attribs, PA4_TYPE_METRIC) &&
		    BGP4_BIT_TEST (attr->attribs, PA4_TYPE_METRIC) &&
		    attn->multiexit == attr->multiexit) {
			/* OK */
		}
		else {
		    BGP4_BIT_RESET (attn->attribs, PA4_TYPE_METRIC);
		}

		if (BGP4_BIT_TEST (attn->attribs, PA4_TYPE_LOCALPREF) &&
		    BGP4_BIT_TEST (attr->attribs, PA4_TYPE_LOCALPREF) &&
		    attn->local_pref == attr->local_pref) {
			/* OK */
		}
		else {
		    BGP4_BIT_RESET (attn->attribs, PA4_TYPE_LOCALPREF);
		}

		if (BGP4_BIT_TEST (attn->attribs, PA4_TYPE_DPA) &&
		    BGP4_BIT_TEST (attr->attribs, PA4_TYPE_DPA) &&
		    attn->dpa.as == attr->dpa.as &&
		    attn->dpa.value == attr->dpa.value) {
			/* OK */
		}
		else {
		    BGP4_BIT_RESET (attn->attribs, PA4_TYPE_DPA);
		}

		if (BGP4_BIT_TEST (attn->attribs, PA4_TYPE_ATOMICAGG) ||
		    BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ATOMICAGG))
        	    BGP4_BIT_SET (attn->attribs, PA4_TYPE_ATOMICAGG);

#ifdef HAVE_IPV6
		if (attn->link_local && attr->link_local &&
		    prefix_compare (attn->link_local, attr->link_local)) {
			/* OK */
		}
		else {
		    if (attn->link_local) {
		        Deref_Prefix (attn->link_local);
		        attn->link_local = NULL;
		    }
		}
		if (attn->nexthop4 && attr->nexthop4 &&
		    prefix_compare (attn->nexthop4, attr->nexthop4)) {
			/* OK */
		}
		else {
		    if (attn->nexthop4) {
		        Deref_Prefix (attn->nexthop4);
		        attn->nexthop4 = NULL;
		    }
		}
#endif /* HAVE_IPV6 */
#ifdef notdef
		if (attn->nexthop2 && attr->nexthop2 &&
		    prefix_compare (attn->nexthop2, attr->nexthop2)) {
			/* OK */
		}
		else {
		    if (attn->nexthop2) {
		        Deref_Prefix (attn->nexthop2);
		        attn->nexthop2 = NULL;
		    }
		}
#endif
    return (attn);
}


/* finalize attn, then return attn */
static bgp_attr_t *
view_aggregate_attr_final (bgp_attr_t *attn, aggregate_t *agg, aspath_t *tail)
{
    assert (attn);

    /* local gateway */
    attn->gateway = agg->attr->gateway;

    if (attn->nexthop2) {
        /* don't inherit nexthop2 */
         Deref_Prefix (attn->nexthop2);
    }
    /* always local */
    attn->nexthop2 = Ref_Prefix (agg->attr->gateway->prefix);

    if (!BGP4_BIT_TEST (attn->attribs, PA4_TYPE_NEXTHOP)) {
        attn->nexthop = Ref_Prefix (agg->attr->nexthop);
        BGP4_BIT_SET (attn->attribs, PA4_TYPE_NEXTHOP);
    }

    /* ignore aggregator of contributing route */
    if (BGP4_BIT_TEST (attn->attribs, PA4_TYPE_AGGREGATOR)) {
        Deref_Prefix (attn->aggregator.prefix);
        attn->aggregator.prefix = NULL;
        BGP4_BIT_RESET (attn->attribs, PA4_TYPE_AGGREGATOR);
    }

    if (BGP4_BIT_TEST (agg->attr->attribs, PA4_TYPE_AGGREGATOR)) {
        attn->aggregator.as = agg->attr->aggregator.as;
        attn->aggregator.prefix = Ref_Prefix (agg->attr->aggregator.prefix);
        BGP4_BIT_SET (attn->attribs, PA4_TYPE_AGGREGATOR);
    }

    /* ignore communities of contributing route */
    if (BGP4_BIT_TEST (attn->attribs, PA4_TYPE_COMMUNITY)) {
        Delete_community (attn->community);
        attn->community = NULL;
        BGP4_BIT_RESET (attn->attribs, PA4_TYPE_COMMUNITY);
    }

#ifdef HAVE_IPV6
    if (attn->link_local == NULL && agg->attr->link_local)
        attn->link_local = Ref_Prefix (agg->attr->link_local);
    if (attn->nexthop4 == NULL && agg->attr->nexthop4)
        attn->nexthop4 = Ref_Prefix (agg->attr->nexthop4);
#endif /* HAVE_IPV6 */

/* CHECK_AGG: */

    if (BIT_TEST (agg->option, BGP_AGGOPT_AS_SET)) {
	if (tail && LL_GetCount (tail)) {
	    attn->aspath = aspath_append (attn->aspath, tail);
            BGP4_BIT_SET (attn->attribs, PA4_TYPE_ASPATH);
	}
        return (attn);
    }

    /* atmoc-aggregate is already set if at least one of routes has */

    if (BGP4_BIT_TEST (attn->attribs, PA4_TYPE_ATOMICAGG))
        return (attn);

    if (BIT_TEST (agg->option, BGP_AGGOPT_SUMMARY_ONLY) &&
	    tail && LL_GetCount (tail)) {
	/* as path information lost, so set atomic-aggregate */
        BGP4_BIT_SET (attn->attribs, PA4_TYPE_ATOMICAGG);
    }

    return (attn);
}


static aggregate_t *
view_check_aggregate (view_t * view, prefix_t *prefix, int withdraw)
{
    aggregate_t *agg;
    radix_node_t *radix_node, *node;
    bgp_route_head_t *bgp_route_head = NULL;
    bgp_attr_t *attr = NULL;
    aspath_t *tail = NULL;
    int n = 0;

    assert (view);
    assert (prefix);
    assert (view->family == prefix->family); /* XXX */

    node = radix_search_best (view->agg_tree, prefix);
    if (node == NULL || /* I'm not sure ... XXX */
	prefix_compare (prefix, node->prefix)) {
	return (NULL);
    }

    agg = RADIX_DATA_GET (node, aggregate_t);
    assert (agg);

    trace (TR_TRACE, view->trace,
	   "AGG contribute: %s -> %s (%s)\n",
	   prefix_toax (prefix),
	   prefix_toax (agg->prefix),
	   (withdraw) ? "withdraw" : "announce");

    if (withdraw) {
        bgp_route_head_t *route_head;
	if (agg->route == NULL)
	    return (NULL);
	radix_node = agg->route->head->radix_node;
	if (radix_node->l || radix_node->r) {
	    /* there is at least one more specific route */
	    /* need recomputation later */
	    goto compute;
	}
	route_head = agg->route->head;
	view_delete_bgp_route (view, agg->route);
	agg->route = NULL;
	/* reseting after deleting the route would be safer */
        if (BIT_TEST (agg->option, BGP_AGGOPT_SUMMARY_ONLY))
	    BGP4_BIT_RESET (route_head->state_bits, VRTS_SUMMARY);
	return (NULL);
    }

    /* the first route causing the aggregate route */
    if (agg->route == NULL) {

	trace (TR_TRACE, view->trace,
		"AGG activate %s triggered by %s\n",
		prefix_toax (agg->prefix),
		prefix_toax (prefix));

	bgp_route_head = view_get_route_head (view, agg->prefix);
	radix_node = bgp_route_head->radix_node;
        if (BIT_TEST (agg->option, BGP_AGGOPT_SUMMARY_ONLY))
	    BGP4_BIT_SET (bgp_route_head->state_bits, VRTS_SUMMARY);
    }
    else {
	radix_node = agg->route->head->radix_node;
    }

    /* view, agg, radix_node, bgp_route_head */
compute:

    RADIX_WALK (radix_node, node) {

        if (n++ == 0) {
            /* the first iteration is aggregate route itself */
        }
        else {
	    bgp_route_head_t *route_head;
	    bgp_route_t *route;

	    route_head = RADIX_DATA_GET (node, bgp_route_head_t);
	    route = (route_head->imported) ? route_head->imported :
	        route_head->active;
	    trace (TR_TRACE, view->trace,
	           "AGG merge attr for %s from %s\n",
	           prefix_toax (agg->prefix),
	           prefix_toax (route_head->prefix));

	    if (attr == NULL) {
	        attr = bgp_copy_attr (route->attr);
	        attr->type = PROTO_BGP;
	    }
	    else {
		if (tail == NULL)
		    tail = New_ASPATH ();
	        attr = view_aggregate_attr_merge (attr, route->attr, agg, 
						  tail);

	        if (route->aggregated)
		    RADIX_WALK_BREAK;
	    }
	}
    }
    RADIX_WALK_END;

    attr = view_aggregate_attr_final (attr, agg, tail);
    if (tail) Delete_ASPATH (tail);

    trace (TR_TRACE, view->trace,
	   "AGG merge result for %s\n",
	   prefix_toax (agg->prefix));
    bgp_trace_attr2 (attr, 0, view->trace);
    if (agg->route == NULL) {
	assert (bgp_route_head);
        agg->route = view_add_bgp_route (view, bgp_route_head, attr, BGP_PREF);
    }
    else {
        agg->route =
	    bgp_implicit_withdraw (view, agg->route, attr, BGP_PREF);
    }
    agg->route->aggregated = 1;
    assert (attr);
    bgp_deref_attr (attr);
    return (agg);
}


/* public */
bgp_route_t *
bgp_add_route (view_t * view, prefix_t * prefix, bgp_attr_t * attr, u_int pref)
{
    bgp_route_t *bgp_route;
    bgp_route_head_t *bgp_route_head;

    assert (view);
    assert (prefix);
    assert (attr);
    if (view->family != prefix->family) {
	trace (TR_ERROR, view->trace,
	       "Not Accept: %s pref %d proto %s\n",
	       prefix_toax (prefix),
	       pref, proto2string (attr->type));
	return (NULL);
    }

    assert (pthread_mutex_trylock (&view->mutex_lock));
    bgp_route_head = view_get_route_head (view, prefix);

    if (attr->type != PROTO_BGP) {	/* extern (imported) route */

	if (bgp_route_head->active &&
	    bgp_route_head->active->pref < pref) {
	    trace (TR_ERROR, view->trace,
	           "Worse than Active?: %s pref %d proto %s (reject)\n",
		   prefix_toax (prefix),
		   bgp_route_head->active->pref,
		   proto2string (bgp_route_head->active->attr->type));
	    bgp_route_head->imported = NULL;
	    return (NULL);
	}

	/* this follows the decision in the RIB */

	if (bgp_route_head->imported)
	    Delete_Bgp_Route (bgp_route_head->imported);
	New_Bgp_Route (bgp_route_head, attr, pref);

	view_add_change_list (view, bgp_route_head, 0);
	return (bgp_route_head->imported);
    }

    /* check to see if we already got this route from the gateway 
     *  if we have, then this is an implicit withdraw of the old route and
     *  we add the new route */
    LL_Iterate (bgp_route_head->ll_routes, bgp_route) {

	if (bgp_route->attr->gateway == attr->gateway)
	    return (bgp_implicit_withdraw (view, bgp_route, attr, pref));
	/* set implicit withdraw flag */
	/* we need a special flag because a host due to policy may not
	   accept the new route, then we have to explicitly withdraw this
	   prefix
	 */
    }

    return (view_add_bgp_route (view, bgp_route_head, attr, pref));
}


/* public */
int
bgp_del_route (view_t * view, prefix_t * prefix, bgp_attr_t * attr)
{
    bgp_route_t *bgp_route;
    bgp_route_head_t *bgp_route_head;

    assert (view);
    assert (prefix);
    assert (attr);
    if (view->family != prefix->family) {
	trace (TR_ERROR, view->trace,
	       "Not Accept: %s proto %s\n",
	       prefix_toax (prefix),
	       proto2string (attr->type));
	return (0);
    }

    assert (pthread_mutex_trylock (&view->mutex_lock));

    bgp_route_head = view_find_route_head (view, prefix);
    if (bgp_route_head == NULL) {
	/* peer can not know if his announce was filtered or not */
	trace (TR_TRACE, view->trace, "Delete: %s not found\n",
	       prefix_toax (prefix));
	return (0);
    }

    if (attr->type != PROTO_BGP) {	/* extern (imported) route */

	if (bgp_route_head->imported == NULL ||
	    bgp_route_head->imported->attr->type != attr->type) {
	    trace (TR_TRACE, view->trace,
		   "Delete: %s proto %s not found\n",
		   prefix_toax (prefix), proto2string (attr->type));
	}
	if (bgp_route_head->imported == NULL)
	    return (0);

	if (bgp_route_head->active == NULL) {
	    /* remove route_head first */
	    view_remove_route_head (view, bgp_route_head);
	    /* add the route head to the change list */
	    view_add_change_list (view, bgp_route_head, 1);
	}
	else {
	    Delete_Bgp_Route (bgp_route_head->imported);
	    bgp_route_head->imported = NULL;
	    /* add the route head to the change list */
	    view_add_change_list (view, bgp_route_head, 0);
	}
	return (1);
    }

    LL_Iterate (bgp_route_head->ll_routes, bgp_route) {
	if (bgp_route->attr->gateway == attr->gateway) {
	    view_delete_bgp_route (view, bgp_route);
	    return (1);
	}
    }
    trace (TR_TRACE, view->trace,
	   "Delete: %s not found\n", prefix_toax (prefix));
    return (0);
}


/* public */
int 
view_close (view_t * view)
{

    pthread_mutex_unlock (&view->mutex_lock);
    return (1);
}


/* public */
int 
view_open (view_t * view)
{

    if (pthread_mutex_trylock (&view->mutex_lock) != 0) {
	trace (TR_DEBUG, view->trace, "Going to block in view_open\n");
	pthread_mutex_lock (&view->mutex_lock);
    }
    return (1);
}


/*
 * this is used in finding nexthop for multihop peer
 */
/* public */
prefix_t *
view_find_best_route (view_t * view, prefix_t * prefix)
{
    radix_node_t *radix_node;
    bgp_route_head_t *route_head;
    bgp_attr_t *attr = NULL;

    assert (view);
    assert (prefix);
    radix_node = radix_search_best (view->radix_tree, prefix);
    if (radix_node) {
	assert (radix_node->data);
	route_head = RADIX_DATA_GET (radix_node, bgp_route_head_t);
	assert (route_head);
	if (route_head->active)
	    attr = route_head->active->attr;
	if (route_head->imported)
	    attr = route_head->imported->attr;
	if (attr && BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP))
	    return (attr->nexthop);
    }
    return (NULL);
}


/* 
 * given a prefix, find route node entry
 * return NULL if one does not exist
 */
/* public */
bgp_route_t *
view_find_bgp_active (view_t * view, prefix_t * prefix)
{
    bgp_route_head_t *bgp_route_head;

    assert (view);
    assert (prefix);
    if ((bgp_route_head = view_find_route_head (view, prefix)) != NULL)
	return (bgp_route_head->active);
    return (NULL);
}


#ifdef notdef
/* view_find_bgp_route
 * given a prefix AND gateway, find route node entry
 * return NULL if one does not exist
 */
bgp_route_t *
view_find_bgp_route (view_t * view,
		     prefix_t * prefix, gateway_t * gateway)
{
    bgp_route_head_t *bgp_route_head;
    bgp_route_t *bgp_route;

    assert (view);
    assert (prefix);
    bgp_route_head = view_get_route_head (view, prefix);

    LL_Iterate (bgp_route_head->ll_routes, bgp_route) {
	if (bgp_route->attr->gateway == gateway) {
	    return (bgp_route);
	}
    }

    return (NULL);
}
#endif


static int
bgp_keep_it (prefix_t *prefix, int type)
{
    assert (type != PROTO_BGP);
    /* network ... */
    if ((type == PROTO_CONNECTED || type == PROTO_STATIC) &&
	check_bgp_networks (prefix))
	return (1);
    /* redistribute */
    if (BIT_TEST (MRT->redist[type], (1 << PROTO_BGP)))
	return (1);
    return (0);
}


/* public */
int
bgp_update_route (prefix_t * prefix, generic_attr_t * new,
		  generic_attr_t * old, int pref)
{
    assert (prefix);

    /*
     * BGP dones't need to hold routes that will not redistibute
     */
    if (new && !bgp_keep_it (prefix, new->type)) {
	new = NULL;
    }
    if (old && !bgp_keep_it (prefix, old->type)) {
	old = NULL;
    }
    if (new == NULL && old == NULL)
	return (0);

    /* in case the rib wants to delete them */
    Ref_Prefix (prefix);
    if (new)
	Ref_Generic_Attr (new);
    if (old)
	Ref_Generic_Attr (old);

    schedule_event2 ("bgp_import",
		     BGP->schedule, bgp_import, 4, prefix, new, old, pref);
    return (1);
}


/* public */
view_t *
New_View (u_int maxbitlen)
{
    view_t *view = New (view_t);

    view->ll_with_routes = LL_Create (0);
    view->ll_ann_routes = LL_Create (0);
    view->agg_tree = New_Radix (maxbitlen);

    /* XXX I know this is a bad way */
    view->family = AF_INET;
#ifdef HAVE_IPV6
    if (maxbitlen == 128) {
	view->family = AF_INET6;
    }
#endif	/* HAVE_IPV6 */

    /*if (LL_ATTR_MEMORY == NULL)
       init_attr_memory (); */

    pthread_mutex_init (&view->mutex_lock, NULL);
    view->radix_tree = New_Radix (maxbitlen);
    return (view);
}


/*
 * withdraw all routes sent by a peer and reset peer bit
 * on all route given to peer when peer leaves established state
 */
/* public */
void
view_delete_peer (view_t * view, bgp_peer_t * peer)
{
    bgp_route_head_t *rt_head;
    bgp_route_t *route;
    LINKED_LIST *ll_route_with;

    assert (view);
    assert (peer);
    view_open (view);

    ll_route_with = LL_Create (0);

    /* find all routes announced by this peer */
    VIEW_RADIX_WALK (view, rt_head) {
	/* we may need to check family? */
	LL_Iterate (rt_head->ll_routes, route) {
	    if (route->attr->gateway == peer->gateway)
		LL_Add (ll_route_with, route);
	    /* reset announce bit */
	    if (view->viewno == peer->view)
	        BIT_RESET (rt_head->peer_mask, peer->bit);
	}
    }
    VIEW_RADIX_WALK_END;

    LL_Iterate (ll_route_with, route) {
	view_delete_bgp_route (view, route);
    }
    LL_Destroy (ll_route_with);
    bgp_process_changes (view);
    view_close (view);
}


/*
 * when a peer is first established, run through view and build
 * BGP updates for all routes per policy
 */
/* public */
void
bgp_establish_peer (bgp_peer_t * peer)
{
    bgp_route_head_t *head;
    bgp_route_t *route;
    LINKED_LIST *ll_ann, *ll_buckets;
    update_bucket_t *bucket;
    view_t *view;

    assert (peer);
    ll_ann = LL_Create (0);

    view = BGP->views[peer->view];
    view_open (view);

    VIEW_RADIX_WALK (view, head) {

	route = head->active;
	if (head->imported /*&& (
		head->imported->attr->type == PROTO_CONNECTED ||
		head->imported->attr->type == PROTO_STATIC)*/) {
	    /* XXX neglect IGP regardless of weight */
	    route = head->imported;
	}

	if (route && bgp_policy (head->prefix, route->attr, peer, 1)) {
	    if (!BIT_TEST (head->state_bits, VRTS_SUPPRESS)) {
	        BIT_SET (head->peer_mask, peer->bit);	/* set announce bit */
		LL_Add (ll_ann, route);
	    }
	    else {
		/* must not be announced */
	        assert (BIT_TEST (head->peer_mask, peer->bit) == 0);
	    }
	}
    }
    VIEW_RADIX_WALK_END;

    ll_buckets = build_update_buckets (ll_ann);

    LL_Iterate (ll_buckets, bucket) {
	view_announce_peer (peer, NULL, bucket->ll_prefix, bucket->attr);
    }
    LL_Destroy (ll_ann);
    LL_Destroy (ll_buckets);

    /* In case that there are somthing changed already */
    LL_ClearFn (view->ll_with_routes, 0);
    LL_ClearFn (view->ll_ann_routes, 0);
    view_close (view);
}


/* public */
int 
show_view (uii_connection_t * uii, int view)
{
    if (view < 0 || view > MAX_BGP_VIEWS || BGP->views[view] == NULL) {
	config_notice (TR_ERROR, uii, 
		       "invalid or unconfigured view %d\r\n", view);
	return (-1);
    }


    pthread_mutex_lock (&BGP->mutex_lock);
    view_open (BGP->views[view]);
    bgp_dump_view (uii, BGP->views[view], NULL, NULL);
    view_close (BGP->views[view]);
    pthread_mutex_unlock (&BGP->mutex_lock);
    return (1);
}


static int 
trace_bgp_view_op (uii_connection_t * uii, char *s, int op)
{
    int i;
    view_t *view;

    if (strcasecmp (s, "*") == 0) {
        for (i = 0; i < MAX_BGP_VIEWS; i++) {
	    if ((view = BGP->views[i]) != NULL) {
	        view_open (view);
        	set_trace (view->trace, op, TR_ALL, NULL);
		view_close (view);
	    }
	}
    }
    else if (isdigit (*s)) {
	i = atoi (s);
	if (i < 0 || i > MAX_BGP_VIEWS ||
	       (view = BGP->views[i]) == NULL) {
	    config_notice (TR_ERROR, uii, 
		       "invalid or unconfigured view %d\r\n", i);
	    Delete (s);
	    return (-1);
	}
	else {
	    view_open (view);
            set_trace (view->trace, op, TR_ALL, NULL);
	    view_close (view);
	}
    }
    else if (strcasecmp (s, "inet") == 0) {
        for (i = 0; i < MAX_BGP_VIEWS; i++) {
	    if ((view = BGP->views[i]) != NULL &&
		 view->family == AF_INET) {
	        view_open (view);
        	set_trace (view->trace, op, TR_ALL, NULL);
		view_close (view);
	    }
	}
    }
#ifdef HAVE_IPV6
    else if (strcasecmp (s, "inet6") == 0) {
        for (i = 0; i < MAX_BGP_VIEWS; i++) {
	    if ((view = BGP->views[i]) != NULL &&
		 view->family == AF_INET6) {
	        view_open (view);
        	set_trace (view->trace, op, TR_ALL, NULL);
		view_close (view);
	    }
	}
    }
#endif /* HAVE_IPV6 */
    else {
	config_notice (TR_ERROR, uii, 
		       "invalid or unconfigured view %s\r\n", s);
        Delete (s);
        return (-1);
    }
    Delete (s);
    return (1);
}

int 
trace_bgp_view (uii_connection_t * uii, char *s)
{
    return (trace_bgp_view_op (uii, s, TRACE_ADD_FLAGS));
}

int 
no_trace_bgp_view (uii_connection_t * uii, char *s)
{
    return (trace_bgp_view_op (uii, s, TRACE_DEL_FLAGS));
}

